# -*- coding: utf-8 -*-
################################################################################
#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
################################################################################
"""
Author: Drift
Email:  wutuobang@baidu.com
Date:   2020/04/22 18:45
"""

from .register import register
from x2paddle.core.util import *


def upsample_shape(input_shapes, scale):
    """

    :param input_shapes:
    :param scale:
    :return:
    """
    assert len(input_shapes) == 1, "not valid input shape for upsample layer"
    assert type(scale) is int

    input_shape = input_shapes[0]
    new_h = scale * input_shape[2]
    new_w = scale * input_shape[3]

    output_shape = [input_shape[0], input_shape[1], new_h, new_w]
    return [output_shape]


def upsample_layer(inputs, scale, input_shape=None, name=None):
    """

    :param inputs:
    :param scale:
    :param input_shape:
    :param name:
    :return:
    """
    x = inputs[0]
    out = fluid.layers.resize_nearest(
        x, align_corners=False, scale=scale, name=name)

    return out


def upsample_weights(name, data=None):
    """

    :param name:
    :param data:
    :return:
    """
    weights_name = []
    return weights_name


register(
    kind='Upsample',
    shape=upsample_shape,
    layer=upsample_layer,
    weights=upsample_weights)



#   Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .adaptive_pool2d_fuser import AdaptivePool2dFuser
from .adaptive_pool2d_fuse_pass import AdaptivePool2dFusePass
from .batchnorm2d_fuser import BatchNorm2dFuser
from .batchnorm2d_fuse_pass import BatchNorm2dFusePass
from .constant_fuser import ConstantFuser
from .constant_fuse_pass import ConstantFusePass
from .dropout_fuser import DropoutFuser
from .dropout_fuse_pass import DropoutFusePass
from .fc_fuser import FcFuser
from .fc_fuse_pass import FcFusePass
from .interpolate_bilinear_fuser import InterpolateBilinearFuser
from .interpolate_bilinear_fuse_pass import InterpolateBilinearFusePass
from .reshape_fuser import ReshapeFuser
from .reshape_fuse_pass import ReshapeFusePass



#   Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from x2paddle.optimizer.pytorch_optimizer.pattern_matcher import FuseBase
from x2paddle.core.program import PaddleGraph, PaddleLayer
from x2paddle.core.util import *


class InterpolateBilinearFuser(FuseBase):
    def __init__(self):
        super(InterpolateBilinearFuser, self).__init__(graph_type="dygraph")
        import torch
        torch_version = torch.__version__
        torch_version_part = torch_version.split(".")
        if int(torch_version_part[0]) == 1 and int(torch_version_part[1]) > 5:
            self.version_gt_150 = True
        else:
            self.version_gt_150 = False

    def build_pattern(self):
        """ 描述需要替换的双线性插值图结构。
        interpolate_bilinear层模式python实现代码示例:
            x3016 = fluid.layers.shape(input=x3005)
            x3016 = len(x3016)
            x3017 = x3016 - 2
            x3018 = []
            for _x3020 in range(x3017):
                x3018.append(None)
            x3021 = (x3005, x8, None, None)
            x3022 = fluid.layers.shape(input=x3005)
            x3022 = len(x3022)
            x3023 = x3022 == 3
            if x3023 :
                raise RaiseException('Exception')
                x3024 = None
            else:
                x3026 = fluid.layers.shape(input=x3005)
                x3026 = len(x3026)
                x3027 = x3026 == 4
                if x3027 :
                    x3044, x3045, x3046, x3047 = x3021
                    x3048 = x3045 is None
                    if x3048 :
                        x3051 = x3046 is None
                        x3049 = x3051
                        x3050 = x3045
                    else:
                        x3052 = x3045
                        x3049 = False
                        x3050 = x3052
                    if x3049 :
                        raise RaiseException('Exception')
                    x3055 = x3050 is not None
                    if x3055 :
                        x3058 = x3050
                        x3059 = x3046 is not None
                        x3056 = x3059
                        x3057 = x3058
                    else:
                        x3056 = False
                        x3057 = x3050
                    if x3056 :
                        raise RaiseException('Exception')
                        x3060 = None
                        x3061 = None
                    else:
                        x3060 = x3046
                        x3061 = x3057
                    x3063 = x3060 is not None
                    if x3063 :
                        x3065 = x3060
                        x3066 = len(x3065)
                        x3067 = x3066 != 2
                        if x3067 :
                            raise RaiseException('Exception')
                        x3064 = x3065
                    else:
                        x3064 = x3060
                    x3070 = x3061 is not None
                    if x3070 :
                        x3072 = x3061
                        x3071 = x3072
                    else:
                        x3071 = None
                    if x3070 :
                        x3073 = x3071
                    else:
                        x3074 = x3064 is not None
                        if x3074 :
                            x3076 = x3064
                            x3075 = x3076
                        else:
                            raise RaiseException('Exception')
                            x3075 = None
                        x3078 = x3047 is None
                        if x3078 :
                            x3080 = len(x3075)
                            x3081 = x3080 > 0
                            x3086 = 0
                            for x3083 in range(2147483647):
                                x3087 = x3075[x3086]
                                x3088 = math.floor(x3087)
                                x3089 = x3088 != x3087
                                if x3089 :
                                    x3090 = False
                                    x3091 = x3089
                                else:
                                    x3090 = None
                                    x3091 = None
                                if x3089 :
                                    x3092 = x3090
                                    x3093 = x3091
                                else:
                                    x3092 = True
                                    x3093 = x3089
                                x3094 = x3086 + 1
                                x3095 = x3094 < x3080
                                x3096 = x3095 and x3092
                                x3082 = x3093
                                x3083 = x3094
                            if x3082 :
                                import warnings
                                warnings.warn('The default behavior for interpolate/upsample with float scale_factor will change in 1.6.0 to align with other frameworks/libraries, and use scale_factor directly, instead of relying on the computed output size. If you wish to keep the old behavior, please set recompute_scale_factor=True. See the documentation of nn.Upsample for details. ', stacklevel=2)
                        x3099 = []
                        for _x3101 in range(2):
                            x3102 = _x3101 + 2
                            x3103 = fluid.layers.shape(x3044)[x3102]
                            x3104 = float(x3103)
                            x3105 = x3075[_x3101]
                            x3106 = x3104 * x3105
                            x3107 = math.floor(x3106)
                            x3099.append(x3107)
                        x3073 = x3099
                    x3108 = x3018[0]
                    x3109 = x3018[1]
                    x3073_isinstance = isinstance(x3073, paddle.fluid.Variable)
                    if x3073_isinstance :
                        x3073 = x3073.numpy().tolist()
                    assert x3108 == x3109, 'The x3108 must be x3109!'
                    x3110 = paddle.nn.functional.interpolate(x=x3005, size=x3073, scale_factor=x3108, align_corners=False, align_mode=0)
                    x3028 = x3110
                else:
                    x3111 = fluid.layers.shape(input=x3005)
                    x3111 = len(x3111)
                    x3112 = x3111 == 5
                    if x3112 :
                        raise RaiseException('Exception')
                    else:
                        raise RaiseException('Exception')
                    x3028 = None
                x3024 = x3028
        """

        def gen_name(id):
            return "x" + str(id)

        if self.version_gt_150:
            self.pattern.add_layer(
                "fluid.layers.shape",
                inputs={"input": "interpolate-input-0"},
                outputs=[gen_name(9)])
            self.pattern.add_layer(
                "prim.len",
                inputs={"input": gen_name(9)},
                outputs=[gen_name(9)])
            self.pattern.add_layer(
                "prim.sub",
                inputs={"x": gen_name(9)},
                outputs=[gen_name(10)],
                y=2)
            self.pattern.add_layer(
                "prim.list", inputs={}, outputs=[gen_name(11)])
            self.pattern.add_layer(
                "prim.loop",
                inputs={"input": gen_name(10)},
                outputs=[gen_name(12.1), gen_name(12.2)])
            loop_layer = self.pattern.layers[list(self.pattern.layers.keys())[
                -1]]
            pattern_block = PaddleGraph(loop_layer, graph_type="dygraph")
            pattern_block.add_layer(
                "prim.append",
                inputs={"list": gen_name(11)},
                outputs=[],
                element=None)
            loop_layer.inputs["input-0"] = gen_name(11)
            loop_layer.add_block(pattern_block)
            self.pattern.add_layer(
                "prim.tuple",
                inputs={
                    "input0": "interpolate-input-0",
                    "input1": "interpolate-input-1",
                },
                outputs=[gen_name(13)],
                input2=None,
                input3=None)
            self.pattern.add_layer(
                "fluid.layers.shape",
                inputs={"input": "interpolate-input-0"},
                outputs=[gen_name(14)])
            self.pattern.add_layer(
                "prim.len",
                inputs={"input": gen_name(14)},
                outputs=[gen_name(14)])
            self.pattern.add_layer(
                "prim.eq",
                inputs={"x": gen_name(14)},
                outputs=[gen_name(15)],
                y=3)
            self.pattern.add_layer(
                "prim.if",
                inputs={"input": gen_name(15)},
                outputs=[gen_name(16)])
            if_layer1 = self.pattern.layers[list(self.pattern.layers.keys())[
                -1]]
            pattern_block = PaddleGraph(if_layer1, graph_type="dygraph")
            pattern_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(17)],
                input="Exception")
            pattern_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(16)], input=None)
            if_layer1.add_block(pattern_block)
            pattern_block = PaddleGraph(if_layer1, graph_type="dygraph")
            pattern_block.add_layer(
                "fluid.layers.shape",
                inputs={"input": "interpolate-input-0"},
                outputs=[gen_name(18)])
            pattern_block.add_layer(
                "prim.len",
                inputs={"input": gen_name(18)},
                outputs=[gen_name(18)])
            pattern_block.add_layer(
                "prim.eq",
                inputs={"x": gen_name(18)},
                outputs=[gen_name(19)],
                y=4)
            pattern_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(19)},
                outputs=[gen_name(20)])
            if_layer2 = pattern_block.layers[list(pattern_block.layers.keys())[
                -1]]
            pattern_block_block = PaddleGraph(if_layer2, graph_type="dygraph")
            pattern_block_block.add_layer(
                "prim.tuple_unpack",
                inputs={"input": gen_name(13)},
                outputs=[
                    gen_name(34), gen_name(35), gen_name(36), gen_name(37)
                ])
            pattern_block_block.add_layer(
                "prim.is",
                inputs={"x": gen_name(35)},
                outputs=[gen_name(38)],
                y=None)
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(38)},
                outputs=[gen_name(39), gen_name(40)])
            if_layer3 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer3, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.is",
                inputs={"x": gen_name(36)},
                outputs=[gen_name(41)],
                y=None)
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(41)},
                outputs=[gen_name(39)])
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(35)},
                outputs=[gen_name(40)])
            if_layer3.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer3, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(35)},
                outputs=[gen_name(42)])
            pattern_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(39)], input=False)
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(35)},
                outputs=[gen_name(40)])
            if_layer3.add_block(pattern_block_block_block)
            if_layer3.inputs.update({
                "input-0": gen_name(36),
                'input-1': gen_name(35),
                'input-2': gen_name(35),
            })
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(39)},
                outputs=[gen_name(43)])
            if_layer4 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer4, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(44)],
                input="Exception")
            if_layer4.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer4, graph_type="dygraph")
            if_layer4.add_block(pattern_block_block_block)
            pattern_block_block.add_layer(
                "prim.isnot",
                inputs={"x": gen_name(40)},
                outputs=[gen_name(45)],
                y=None)
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(45)},
                outputs=[gen_name(46), gen_name(47)])
            if_layer5 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer5, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(40)},
                outputs=[gen_name(48)])
            pattern_block_block_block.add_layer(
                "prim.isnot",
                inputs={"x": gen_name(36)},
                outputs=[gen_name(49)],
                y=None)
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(49)},
                outputs=[gen_name(46)])
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(48)},
                outputs=[gen_name(47)])
            if_layer5.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer5, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(46)], input=False)
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(40)},
                outputs=[gen_name(47)])
            if_layer5.add_block(pattern_block_block_block)
            if_layer5.inputs.update({
                "input-0": gen_name(40),
                "input-1": gen_name(36),
                "input-3": gen_name(40)
            })
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(46)},
                outputs=[gen_name(50), gen_name(51)])
            if_layer6 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer6, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(52)],
                input="Exception")
            pattern_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(50)], input=None)
            pattern_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(51)], input=None)
            if_layer6.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer6, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(36)},
                outputs=[gen_name(50)])
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(47)},
                outputs=[gen_name(51)])
            if_layer6.add_block(pattern_block_block_block)
            if_layer6.inputs.update({
                "input-0": gen_name(36),
                "input-1": gen_name(47)
            })
            pattern_block_block.add_layer(
                "prim.isnot",
                inputs={"x": gen_name(50)},
                outputs=[gen_name(53)],
                y=None)
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(53)},
                outputs=[gen_name(54)])
            if_layer7 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer7, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(50)},
                outputs=[gen_name(55)])
            pattern_block_block_block.add_layer(
                "prim.len",
                inputs={"input": gen_name(55)},
                outputs=[gen_name(56)])
            pattern_block_block_block.add_layer(
                "prim.ne",
                inputs={"x": gen_name(56)},
                outputs=[gen_name(57)],
                y=2)
            pattern_block_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(57)},
                outputs=[gen_name(58)])
            if_layer8 = pattern_block_block_block.layers[list(
                pattern_block_block_block.layers.keys())[-1]]
            pattern_block_block_block_block = PaddleGraph(
                if_layer8, graph_type="dygraph")
            pattern_block_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(59)],
                input="Exception")
            if_layer8.add_block(pattern_block_block_block_block)
            pattern_block_block_block_block = PaddleGraph(
                if_layer8, graph_type="dygraph")
            if_layer8.add_block(pattern_block_block_block_block)
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(55)},
                outputs=[gen_name(54)])
            if_layer7.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer7, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(50)},
                outputs=[gen_name(54)])
            if_layer7.add_block(pattern_block_block_block)
            if_layer7.inputs.update({
                "input-0": gen_name(50),
                "input-1": gen_name(50)
            })
            pattern_block_block.add_layer(
                "prim.isnot",
                inputs={"x": gen_name(51)},
                outputs=[gen_name(60)],
                y=None)
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(60)},
                outputs=[gen_name(61)])
            if_layer9 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer9, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(51)},
                outputs=[gen_name(62)])
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(62)},
                outputs=[gen_name(61)])
            if_layer9.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer9, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.isnot",
                inputs={"x": gen_name(54)},
                outputs=[gen_name(64)],
                y=None)
            pattern_block_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(64)},
                outputs=[gen_name(65)])
            if_layer11 = pattern_block_block_block.layers[list(
                pattern_block_block_block.layers.keys())[-1]]
            pattern_block_block_block_block = PaddleGraph(
                if_layer11, graph_type="dygraph")
            pattern_block_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(54)},
                outputs=[gen_name(66)])
            pattern_block_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(66)},
                outputs=[gen_name(65)])
            if_layer11.add_block(pattern_block_block_block_block)
            pattern_block_block_block_block = PaddleGraph(
                if_layer11, graph_type="dygraph")
            pattern_block_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(67)],
                input="Exception")
            pattern_block_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(65)], input=None)
            if_layer11.add_block(pattern_block_block_block_block)
            if_layer11.inputs.update({"input-0": gen_name(54), })
            pattern_block_block_block.add_layer(
                "prim.is",
                inputs={"x": gen_name(37)},
                outputs=[gen_name(68)],
                y=None)
            pattern_block_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(68)},
                outputs=[gen_name(69)])
            if_layer12 = pattern_block_block_block.layers[list(
                pattern_block_block_block.layers.keys())[-1]]
            pattern_block_block_block_block = PaddleGraph(
                if_layer12, graph_type="dygraph")
            pattern_block_block_block_block.add_layer(
                "prim.len",
                inputs={"input": gen_name(65)},
                outputs=[gen_name(70)])
            pattern_block_block_block_block.add_layer(
                "prim.gt",
                inputs={"x": gen_name(70)},
                outputs=[gen_name(71)],
                y=0)
            pattern_block_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(72)], input=0)
            pattern_block_block_block_block.add_layer(
                "prim.loop",
                inputs={},
                outputs=[gen_name(74), gen_name(75), gen_name(76.1)],
                input=2147483647)
            loop_layer = pattern_block_block_block_block.layers[list(
                pattern_block_block_block_block.layers.keys())[-1]]
            pattern_loop_block = PaddleGraph(loop_layer, graph_type="dygraph")
            pattern_loop_block.add_layer(
                "prim.getitem",
                inputs={"list": gen_name(65),
                        "element": gen_name(72)},
                outputs=[gen_name(74.1)])
            pattern_loop_block.add_layer(
                "prim.floor",
                inputs={"input": gen_name(74.1)},
                outputs=[gen_name(75.1)])
            pattern_loop_block.add_layer(
                "prim.ne",
                inputs={"x": gen_name(75.1),
                        "y": gen_name(74.1)},
                outputs=[gen_name(76)])
            pattern_loop_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(76)},
                outputs=[gen_name(77)])
            if_layer13 = pattern_loop_block.layers[list(
                pattern_loop_block.layers.keys())[-1]]
            pattern_loop_block_block = PaddleGraph(
                if_layer13, graph_type="dygraph")
            pattern_loop_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(77)], input=False)
            if_layer13.add_block(pattern_loop_block_block)
            pattern_loop_block_block = PaddleGraph(
                if_layer13, graph_type="dygraph")
            pattern_loop_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(77)], input=True)
            if_layer13.add_block(pattern_loop_block_block)
            pattern_loop_block.add_layer(
                "prim.add",
                inputs={"x": gen_name(72)},
                outputs=[gen_name(81)],
                y=1)
            pattern_loop_block.add_layer(
                "prim.lt",
                inputs={"x": gen_name(81),
                        "y": gen_name(70)},
                outputs=[gen_name(82)])
            pattern_loop_block.add_layer(
                "prim.and",
                inputs={"x": gen_name(82),
                        "y": gen_name(77)},
                outputs=[gen_name(83)])
            pattern_loop_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(76)},
                outputs=[gen_name(74)])
            pattern_loop_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(81)},
                outputs=[gen_name(75)])
            loop_layer.add_block(pattern_loop_block)
            loop_layer.inputs.update({
                "input-0": gen_name(65),
                "input-1": gen_name(72),
                "input-2": gen_name(72),
                "input-3": gen_name(70)
            })
            pattern_block_block_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(74)},
                outputs=[gen_name(84)])
            if_layer15 = pattern_block_block_block_block.layers[list(
                pattern_block_block_block_block.layers.keys())[-1]]
            pattern_block_block_block_block_block = PaddleGraph(
                if_layer15, graph_type="dygraph")
            pattern_block_block_block_block_block.add_layer(
                "prim.warnings",
                inputs={},
                outputs=[gen_name(85)],
                stacklevel=2,
                input="...")
            if_layer15.add_block(pattern_block_block_block_block_block)
            pattern_block_block_block_block_block = PaddleGraph(
                if_layer15, graph_type="dygraph")
            if_layer15.add_block(pattern_block_block_block_block_block)
            if_layer12.add_block(pattern_block_block_block_block)
            pattern_block_block_block_block = PaddleGraph(
                if_layer12, graph_type="dygraph")
            if_layer12.add_block(pattern_block_block_block_block)
            if_layer12.inputs.update({
                "input-0": gen_name(65),
                "input-1": gen_name(65),
            })
            pattern_block_block_block.add_layer(
                "prim.list", inputs={}, outputs=[gen_name(86)])
            pattern_block_block_block.add_layer(
                "prim.loop",
                inputs={},
                outputs=[gen_name(87), gen_name(88)],
                input=2)
            loop_layer = pattern_block_block_block.layers[list(
                pattern_block_block_block.layers.keys())[-1]]
            pattern_loop_block = PaddleGraph(loop_layer, graph_type="dygraph")
            pattern_loop_block.add_layer(
                "prim.add",
                inputs={"x": gen_name(88)},
                outputs=[gen_name(89)],
                y=2)
            pattern_loop_block.add_layer(
                "prim.shape_dim",
                inputs={"input": gen_name(34),
                        "dim": gen_name(89)},
                outputs=[gen_name(90)])
            pattern_loop_block.add_layer(
                "prim.float",
                inputs={"input": gen_name(90)},
                outputs=[gen_name(91)])
            pattern_loop_block.add_layer(
                "prim.getitem",
                inputs={"list": gen_name(65),
                        "element": gen_name(88)},
                outputs=[gen_name(92)])
            pattern_loop_block.add_layer(
                "prim.mul",
                inputs={"x": gen_name(91),
                        "y": gen_name(92)},
                outputs=[gen_name(93)])
            pattern_loop_block.add_layer(
                "prim.floor",
                inputs={"input": gen_name(93)},
                outputs=[gen_name(94)])
            pattern_loop_block.add_layer(
                "prim.append",
                inputs={"list": gen_name(86),
                        "element": gen_name(94)},
                outputs=[])
            loop_layer.add_block(pattern_loop_block)
            loop_layer.inputs.update({
                "input-0": gen_name(34),
                "input-1": gen_name(65),
                "input-2": gen_name(86)
            })
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(86)},
                outputs=[gen_name(61)])
            if_layer9.add_block(pattern_block_block_block)
            if_layer9.inputs.update({
                "input-0": gen_name(51),
                "input-1": gen_name(54),
                "input-2": gen_name(54),
                "input-3": gen_name(37),
                "input-4": gen_name(34)
            })
            pattern_block_block.add_layer(
                "prim.getitem",
                inputs={"list": gen_name(11)},
                outputs=[gen_name(95)],
                element=0)
            pattern_block_block.add_layer(
                "prim.getitem",
                inputs={"list": gen_name(11)},
                outputs=[gen_name(96)],
                element=1)
            pattern_block_block.add_layer(
                "prim.isinstance",
                inputs={"input": gen_name(61)},
                outputs=["interpolate-input-0_isinstance"],
                cls="paddle.fluid.Variable")
            pattern_block_block.add_layer(
                "prim.if", {"input": "interpolate-input-0_isinstance"},
                outputs=["interpolate-input-0_if1"])
            if_layer_isinstance = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer_isinstance, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.var2list",
                inputs={"input": gen_name(61)},
                outputs=[gen_name(61)])
            if_layer_isinstance.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer_isinstance, graph_type="dygraph")
            if_layer_isinstance.add_block(pattern_block_block_block)
            if_layer_isinstance.inputs["input-0"] = gen_name(61)
            pattern_block_block.add_layer(
                "prim.assert",
                inputs={"key": gen_name(95),
                        "value": gen_name(96)},
                outputs=[gen_name(97) + "_assert"],
                type="eq")
            pattern_block_block.add_layer(
                "paddle.nn.functional.interpolate",
                inputs={
                    "input": "interpolate-input-0",
                    "size": gen_name(61),
                    "scale_factor": gen_name(95)
                },
                outputs=[gen_name(97)],
                align_corners=False,
                align_mode=0)
            pattern_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(97)},
                outputs=[gen_name(20)])
            if_layer2.add_block(pattern_block_block)
            pattern_block_block = PaddleGraph(if_layer2, graph_type="dygraph")
            pattern_block_block.add_layer(
                "fluid.layers.shape",
                inputs={"input": "interpolate-input-0"},
                outputs=[gen_name(98)])
            pattern_block_block.add_layer(
                "prim.len",
                inputs={"input": gen_name(98)},
                outputs=[gen_name(98)])
            pattern_block_block.add_layer(
                "prim.eq",
                inputs={"x": gen_name(98)},
                outputs=[gen_name(99)],
                y=5)
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(99)},
                outputs=[gen_name(100)])
            if_layer16 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer16, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(101)],
                input="Exception")
            if_layer16.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer16, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(102)],
                input="Exception")
            if_layer16.add_block(pattern_block_block_block)
            pattern_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(20)], input=None)
            if_layer2.add_block(pattern_block_block)
            if_layer2.inputs.update({
                "input-0": gen_name(13),
                "input-1": gen_name(13),
                "input-2": "interpolate-input-0",
                "input-3": gen_name(11),
                "input-5": gen_name(11),
            })
            pattern_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(20)},
                outputs=[gen_name(16)])
            if_layer1.add_block(pattern_block)
            if_layer1.inputs.update({
                "input-2": "interpolate-input-0",
                "input-4": gen_name(13),
                "input-7": gen_name(11),
                "input-9": gen_name(11),
                "input-11": "interpolate-input-0",
                "input-12": "interpolate-input-0",
            })
            self.pattern.build(inputs={
                "input-0": "interpolate-input-0",
                "input-1": "interpolate-input-1"
            })
        else:
            self.pattern.add_layer(
                "fluid.layers.shape",
                inputs={"input": "interpolate-input-0"},
                outputs=[gen_name(9)])
            self.pattern.add_layer(
                "prim.len",
                inputs={"input": gen_name(9)},
                outputs=[gen_name(9)])
            self.pattern.add_layer(
                "prim.sub",
                inputs={"x": gen_name(9)},
                outputs=[gen_name(10)],
                y=2)
            self.pattern.add_layer(
                "prim.list", inputs={}, outputs=[gen_name(11)])
            self.pattern.add_layer(
                "prim.loop",
                inputs={"input": gen_name(10)},
                outputs=[gen_name(12.1), gen_name(12.2)])
            loop_layer = self.pattern.layers[list(self.pattern.layers.keys())[
                -1]]
            pattern_block = PaddleGraph(loop_layer, graph_type="dygraph")
            pattern_block.add_layer(
                "prim.append",
                inputs={"list": gen_name(11)},
                outputs=[],
                element=None)
            loop_layer.inputs["input-0"] = gen_name(11)
            loop_layer.add_block(pattern_block)
            self.pattern.add_layer(
                "prim.tuple",
                inputs={
                    "input0": "interpolate-input-0",
                    "input1": "interpolate-input-1",
                },
                outputs=[gen_name(13)],
                input2=None,
                input3=None)
            self.pattern.add_layer(
                "fluid.layers.shape",
                inputs={"input": "interpolate-input-0"},
                outputs=[gen_name(14)])
            self.pattern.add_layer(
                "prim.len",
                inputs={"input": gen_name(14)},
                outputs=[gen_name(14)])
            self.pattern.add_layer(
                "prim.eq",
                inputs={"x": gen_name(14)},
                outputs=[gen_name(15)],
                y=3)
            self.pattern.add_layer(
                "prim.if",
                inputs={"input": gen_name(15)},
                outputs=[gen_name(16)])
            if_layer1 = self.pattern.layers[list(self.pattern.layers.keys())[
                -1]]
            pattern_block = PaddleGraph(if_layer1, graph_type="dygraph")
            pattern_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(17)],
                input="Exception")
            pattern_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(16)], input=None)
            if_layer1.add_block(pattern_block)
            pattern_block = PaddleGraph(if_layer1, graph_type="dygraph")
            pattern_block.add_layer(
                "fluid.layers.shape",
                inputs={"input": "interpolate-input-0"},
                outputs=[gen_name(18)])
            pattern_block.add_layer(
                "prim.len",
                inputs={"input": gen_name(18)},
                outputs=[gen_name(18)])
            pattern_block.add_layer(
                "prim.eq",
                inputs={"x": gen_name(18)},
                outputs=[gen_name(19)],
                y=4)
            pattern_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(19)},
                outputs=[gen_name(20)])
            if_layer2 = pattern_block.layers[list(pattern_block.layers.keys())[
                -1]]
            pattern_block_block = PaddleGraph(if_layer2, graph_type="dygraph")
            pattern_block_block.add_layer(
                "prim.tuple_unpack",
                inputs={"input": gen_name(13)},
                outputs=[
                    gen_name(34), gen_name(35), gen_name(36), gen_name(37)
                ])
            pattern_block_block.add_layer(
                "prim.is",
                inputs={"x": gen_name(35)},
                outputs=[gen_name(38)],
                y=None)
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(38)},
                outputs=[gen_name(39), gen_name(40)])
            if_layer3 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer3, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.is",
                inputs={"x": gen_name(36)},
                outputs=[gen_name(41)],
                y=None)
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(41)},
                outputs=[gen_name(39)])
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(35)},
                outputs=[gen_name(40)])
            if_layer3.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer3, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(35)},
                outputs=[gen_name(42)])
            pattern_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(39)], input=False)
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(35)},
                outputs=[gen_name(40)])
            if_layer3.add_block(pattern_block_block_block)
            if_layer3.inputs.update({
                "input-0": gen_name(36),
                'input-1': gen_name(35),
                'input-2': gen_name(35),
            })
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(39)},
                outputs=[gen_name(43)])
            if_layer4 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer4, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(44)],
                input="Exception")
            if_layer4.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer4, graph_type="dygraph")
            if_layer4.add_block(pattern_block_block_block)
            pattern_block_block.add_layer(
                "prim.isnot",
                inputs={"x": gen_name(40)},
                outputs=[gen_name(45)],
                y=None)
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(45)},
                outputs=[gen_name(46), gen_name(47)])
            if_layer5 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer5, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(40)},
                outputs=[gen_name(48)])
            pattern_block_block_block.add_layer(
                "prim.isnot",
                inputs={"x": gen_name(36)},
                outputs=[gen_name(49)],
                y=None)
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(49)},
                outputs=[gen_name(46)])
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(48)},
                outputs=[gen_name(47)])
            if_layer5.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer5, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(46)], input=False)
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(40)},
                outputs=[gen_name(47)])
            if_layer5.add_block(pattern_block_block_block)
            if_layer5.inputs.update({
                "input-0": gen_name(40),
                "input-1": gen_name(36),
                "input-3": gen_name(40)
            })
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(46)},
                outputs=[gen_name(50), gen_name(51)])
            if_layer6 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer6, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(52)],
                input="Exception")
            pattern_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(50)], input=None)
            pattern_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(51)], input=None)
            if_layer6.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer6, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(36)},
                outputs=[gen_name(50)])
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(47)},
                outputs=[gen_name(51)])
            if_layer6.add_block(pattern_block_block_block)
            if_layer6.inputs.update({
                "input-0": gen_name(36),
                "input-1": gen_name(47)
            })
            pattern_block_block.add_layer(
                "prim.isnot",
                inputs={"x": gen_name(50)},
                outputs=[gen_name(53)],
                y=None)
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(53)},
                outputs=[gen_name(54)])
            if_layer7 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer7, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(50)},
                outputs=[gen_name(55)])
            pattern_block_block_block.add_layer(
                "prim.len",
                inputs={"input": gen_name(55)},
                outputs=[gen_name(56)])
            pattern_block_block_block.add_layer(
                "prim.ne",
                inputs={"x": gen_name(56)},
                outputs=[gen_name(57)],
                y=2)
            pattern_block_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(57)},
                outputs=[gen_name(58)])
            if_layer8 = pattern_block_block_block.layers[list(
                pattern_block_block_block.layers.keys())[-1]]
            pattern_block_block_block_block = PaddleGraph(
                if_layer8, graph_type="dygraph")
            pattern_block_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(59)],
                input="Exception")
            if_layer8.add_block(pattern_block_block_block_block)
            pattern_block_block_block_block = PaddleGraph(
                if_layer8, graph_type="dygraph")
            if_layer8.add_block(pattern_block_block_block_block)
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(55)},
                outputs=[gen_name(54)])
            if_layer7.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer7, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(50)},
                outputs=[gen_name(54)])
            if_layer7.add_block(pattern_block_block_block)
            if_layer7.inputs.update({
                "input-0": gen_name(50),
                "input-1": gen_name(50)
            })
            pattern_block_block.add_layer(
                "prim.isnot",
                inputs={"x": gen_name(51)},
                outputs=[gen_name(60)],
                y=None)
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(60)},
                outputs=[gen_name(61)])
            if_layer9 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer9, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(51)},
                outputs=[gen_name(62)])
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(62)},
                outputs=[gen_name(61)])
            if_layer9.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer9, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(61)], input=None)
            if_layer9.add_block(pattern_block_block_block)
            if_layer9.inputs.update({"input-0": gen_name(51)})
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(60)},
                outputs=[gen_name(63)])
            if_layer10 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer10, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(61)},
                outputs=[gen_name(63)])
            if_layer10.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer10, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.isnot",
                inputs={"x": gen_name(54)},
                outputs=[gen_name(64)],
                y=None)
            pattern_block_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(64)},
                outputs=[gen_name(65)])
            if_layer11 = pattern_block_block_block.layers[list(
                pattern_block_block_block.layers.keys())[-1]]
            pattern_block_block_block_block = PaddleGraph(
                if_layer11, graph_type="dygraph")
            pattern_block_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(54)},
                outputs=[gen_name(66)])
            pattern_block_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(66)},
                outputs=[gen_name(65)])
            if_layer11.add_block(pattern_block_block_block_block)
            pattern_block_block_block_block = PaddleGraph(
                if_layer11, graph_type="dygraph")
            pattern_block_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(67)],
                input="Exception")
            pattern_block_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(65)], input=None)
            if_layer11.add_block(pattern_block_block_block_block)
            if_layer11.inputs.update({"input-0": gen_name(54), })
            pattern_block_block_block.add_layer(
                "prim.is",
                inputs={"x": gen_name(37)},
                outputs=[gen_name(68)],
                y=None)
            pattern_block_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(68)},
                outputs=[gen_name(69)])
            if_layer12 = pattern_block_block_block.layers[list(
                pattern_block_block_block.layers.keys())[-1]]
            pattern_block_block_block_block = PaddleGraph(
                if_layer12, graph_type="dygraph")
            pattern_block_block_block_block.add_layer(
                "prim.len",
                inputs={"input": gen_name(65)},
                outputs=[gen_name(70)])
            pattern_block_block_block_block.add_layer(
                "prim.gt",
                inputs={"x": gen_name(70)},
                outputs=[gen_name(71)],
                y=0)
            pattern_block_block_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(72)], input=0)
            pattern_block_block_block_block.add_layer(
                "prim.loop",
                inputs={},
                outputs=[gen_name(74), gen_name(75), gen_name(76.1)],
                input=2147483647)
            loop_layer = pattern_block_block_block_block.layers[list(
                pattern_block_block_block_block.layers.keys())[-1]]
            pattern_loop_block = PaddleGraph(loop_layer, graph_type="dygraph")
            pattern_loop_block.add_layer(
                "prim.getitem",
                inputs={"list": gen_name(65),
                        "element": gen_name(72)},
                outputs=[gen_name(74.1)])
            pattern_loop_block.add_layer(
                "prim.floor",
                inputs={"input": gen_name(74.1)},
                outputs=[gen_name(75.1)])
            pattern_loop_block.add_layer(
                "prim.ne",
                inputs={"x": gen_name(75.1),
                        "y": gen_name(74.1)},
                outputs=[gen_name(76)])
            pattern_loop_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(76)},
                outputs=[gen_name(77), gen_name(78)])
            if_layer13 = pattern_loop_block.layers[list(
                pattern_loop_block.layers.keys())[-1]]
            pattern_loop_block_block = PaddleGraph(
                if_layer13, graph_type="dygraph")
            pattern_loop_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(77)], input=False)
            pattern_loop_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(76)},
                outputs=[gen_name(78)])
            if_layer13.add_block(pattern_loop_block_block)
            pattern_loop_block_block = PaddleGraph(
                if_layer13, graph_type="dygraph")
            pattern_loop_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(77)], input=None)
            pattern_loop_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(78)], input=None)
            if_layer13.add_block(pattern_loop_block_block)
            if_layer13.inputs.update({"input-0": gen_name(76), })
            pattern_loop_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(76)},
                outputs=[gen_name(79), gen_name(80)])
            if_layer14 = pattern_loop_block.layers[list(
                pattern_loop_block.layers.keys())[-1]]
            pattern_loop_block_block = PaddleGraph(
                if_layer14, graph_type="dygraph")
            pattern_loop_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(77)},
                outputs=[gen_name(79)])
            pattern_loop_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(78)},
                outputs=[gen_name(80)])
            if_layer14.add_block(pattern_loop_block_block)
            pattern_loop_block_block = PaddleGraph(
                if_layer14, graph_type="dygraph")
            pattern_loop_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(79)], input=True)
            pattern_loop_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(76)},
                outputs=[gen_name(80)])
            if_layer14.add_block(pattern_loop_block_block)
            if_layer14.inputs.update({
                "input-0": gen_name(77),
                "input-1": gen_name(78),
                "input-2": gen_name(76)
            })
            pattern_loop_block.add_layer(
                "prim.add",
                inputs={"x": gen_name(72)},
                outputs=[gen_name(81)],
                y=1)
            pattern_loop_block.add_layer(
                "prim.lt",
                inputs={"x": gen_name(81),
                        "y": gen_name(70)},
                outputs=[gen_name(82)])
            pattern_loop_block.add_layer(
                "prim.and",
                inputs={"x": gen_name(82),
                        "y": gen_name(79)},
                outputs=[gen_name(83)])
            pattern_loop_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(80)},
                outputs=[gen_name(74)])
            pattern_loop_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(81)},
                outputs=[gen_name(75)])
            loop_layer.add_block(pattern_loop_block)
            loop_layer.inputs.update({
                "input-0": gen_name(65),
                "input-1": gen_name(72),
                "input-2": gen_name(72),
                "input-3": gen_name(70)
            })
            pattern_block_block_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(74)},
                outputs=[gen_name(84)])
            if_layer15 = pattern_block_block_block_block.layers[list(
                pattern_block_block_block_block.layers.keys())[-1]]
            pattern_block_block_block_block_block = PaddleGraph(
                if_layer15, graph_type="dygraph")
            pattern_block_block_block_block_block.add_layer(
                "prim.warnings",
                inputs={},
                outputs=[gen_name(85)],
                stacklevel=2,
                input="...")
            if_layer15.add_block(pattern_block_block_block_block_block)
            pattern_block_block_block_block_block = PaddleGraph(
                if_layer15, graph_type="dygraph")
            if_layer15.add_block(pattern_block_block_block_block_block)
            if_layer12.add_block(pattern_block_block_block_block)
            pattern_block_block_block_block = PaddleGraph(
                if_layer12, graph_type="dygraph")
            if_layer12.add_block(pattern_block_block_block_block)
            if_layer12.inputs.update({
                "input-0": gen_name(65),
                "input-1": gen_name(65),
            })
            pattern_block_block_block.add_layer(
                "prim.list", inputs={}, outputs=[gen_name(86)])
            pattern_block_block_block.add_layer(
                "prim.loop",
                inputs={},
                outputs=[gen_name(87), gen_name(88)],
                input=2)
            loop_layer = pattern_block_block_block.layers[list(
                pattern_block_block_block.layers.keys())[-1]]
            pattern_loop_block = PaddleGraph(loop_layer, graph_type="dygraph")
            pattern_loop_block.add_layer(
                "prim.add",
                inputs={"x": gen_name(88)},
                outputs=[gen_name(89)],
                y=2)
            pattern_loop_block.add_layer(
                "prim.shape_dim",
                inputs={"input": gen_name(34),
                        "dim": gen_name(89)},
                outputs=[gen_name(90)])
            pattern_loop_block.add_layer(
                "prim.float",
                inputs={"input": gen_name(90)},
                outputs=[gen_name(91)])
            pattern_loop_block.add_layer(
                "prim.getitem",
                inputs={"list": gen_name(65),
                        "element": gen_name(88)},
                outputs=[gen_name(92)])
            pattern_loop_block.add_layer(
                "prim.mul",
                inputs={"x": gen_name(91),
                        "y": gen_name(92)},
                outputs=[gen_name(93)])
            pattern_loop_block.add_layer(
                "prim.floor",
                inputs={"input": gen_name(93)},
                outputs=[gen_name(94)])
            pattern_loop_block.add_layer(
                "prim.append",
                inputs={"list": gen_name(86),
                        "element": gen_name(94)},
                outputs=[])
            loop_layer.add_block(pattern_loop_block)
            loop_layer.inputs.update({
                "input-0": gen_name(34),
                "input-1": gen_name(65),
                "input-2": gen_name(86)
            })
            pattern_block_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(86)},
                outputs=[gen_name(63)])
            if_layer10.add_block(pattern_block_block_block)
            if_layer10.inputs.update({
                "input-0": gen_name(61),
                "input-1": gen_name(54),
                "input-2": gen_name(54),
                "input-3": gen_name(37),
                "input-4": gen_name(34)
            })
            pattern_block_block.add_layer(
                "prim.getitem",
                inputs={"list": gen_name(11)},
                outputs=[gen_name(95)],
                element=0)
            pattern_block_block.add_layer(
                "prim.getitem",
                inputs={"list": gen_name(11)},
                outputs=[gen_name(96)],
                element=1)
            pattern_block_block.add_layer(
                "prim.isinstance",
                inputs={"input": gen_name(63)},
                outputs=["interpolate-input-0_isinstance"],
                cls="paddle.fluid.Variable")
            pattern_block_block.add_layer(
                "prim.if", {"input": "interpolate-input-0_isinstance"},
                outputs=["interpolate-input-0_if1"])
            if_layer_isinstance = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer_isinstance, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.var2list",
                inputs={"input": gen_name(63)},
                outputs=[gen_name(63)])
            if_layer_isinstance.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer_isinstance, graph_type="dygraph")
            if_layer_isinstance.add_block(pattern_block_block_block)
            if_layer_isinstance.inputs["input-0"] = gen_name(63)
            pattern_block_block.add_layer(
                "prim.assert",
                inputs={"key": gen_name(95),
                        "value": gen_name(96)},
                outputs=[gen_name(97) + "_assert"],
                type="eq")
            pattern_block_block.add_layer(
                "paddle.nn.functional.interpolate",
                inputs={
                    "input": "interpolate-input-0",
                    "size": gen_name(63),
                    "scale_factor": gen_name(95)
                },
                outputs=[gen_name(97)],
                align_corners=False,
                align_mode=0)
            pattern_block_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(97)},
                outputs=[gen_name(20)])
            if_layer2.add_block(pattern_block_block)
            pattern_block_block = PaddleGraph(if_layer2, graph_type="dygraph")
            pattern_block_block.add_layer(
                "fluid.layers.shape",
                inputs={"input": "interpolate-input-0"},
                outputs=[gen_name(98)])
            pattern_block_block.add_layer(
                "prim.len",
                inputs={"input": gen_name(98)},
                outputs=[gen_name(98)])
            pattern_block_block.add_layer(
                "prim.eq",
                inputs={"x": gen_name(98)},
                outputs=[gen_name(99)],
                y=5)
            pattern_block_block.add_layer(
                "prim.if",
                inputs={"input": gen_name(99)},
                outputs=[gen_name(100)])
            if_layer16 = pattern_block_block.layers[list(
                pattern_block_block.layers.keys())[-1]]
            pattern_block_block_block = PaddleGraph(
                if_layer16, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(101)],
                input="Exception")
            if_layer16.add_block(pattern_block_block_block)
            pattern_block_block_block = PaddleGraph(
                if_layer16, graph_type="dygraph")
            pattern_block_block_block.add_layer(
                "prim.exception",
                inputs={},
                outputs=[gen_name(102)],
                input="Exception")
            if_layer16.add_block(pattern_block_block_block)
            pattern_block_block.add_layer(
                "prim.equal", inputs={}, outputs=[gen_name(20)], input=None)
            if_layer2.add_block(pattern_block_block)
            if_layer2.inputs.update({
                "input-0": gen_name(13),
                "input-1": gen_name(13),
                "input-2": "interpolate-input-0",
                "input-3": gen_name(11),
                "input-5": gen_name(11),
            })
            pattern_block.add_layer(
                "prim.equal",
                inputs={"input": gen_name(20)},
                outputs=[gen_name(16)])
            if_layer1.add_block(pattern_block)
            if_layer1.inputs.update({
                "input-2": "interpolate-input-0",
                "input-4": gen_name(13),
                "input-7": gen_name(11),
                "input-9": gen_name(11),
                "input-11": "interpolate-input-0",
                "input-12": "interpolate-input-0",
            })
            self.pattern.build(inputs={
                "input-0": "interpolate-input-0",
                "input-1": "interpolate-input-1"
            })

    def insert_new_layer(self, graph, parameters, matches):
        new_layers = self.gen_new_layer(parameters, matches)
        new_layer_id = list(matches.keys())[0]
        graph.layers[new_layer_id] = new_layers[0]
        matches.pop(new_layer_id)
        new_layer_id = list(matches.keys())[0]
        graph.layers[new_layer_id] = new_layers[1]
        block_layer = new_layers[1].blocks[0].layers.pop(
            list(new_layers[1].blocks[0].layers.keys())[-1])
        new_layers[1].blocks[0].layers[new_layer_id + ".0.0"] = block_layer
        matches.pop(new_layer_id)
        new_layer_id = list(matches.keys())[0]
        graph.layers[new_layer_id] = new_layers[2]
        matches.pop(new_layer_id)

    def gen_new_layer(self, parameters, matches):
        layers = list()
        layers_id = list(matches.keys())
        layer = matches[layers_id[6]]
        size = layer.inputs["input1"]
        layer = matches[layers_id[92]]
        layer.inputs["input"] = size
        layers.append(layer)
        layer = matches[layers_id[93]]
        block_layer = layer.blocks[0].layers[list(layer.blocks[0].layers.keys())
                                             [0]]
        block_layer.inputs["input"] = size
        block_layer.outputs[0] = size
        layer.inputs["input-0"] = size
        layers.append(layer)
        layer = matches[layers_id[-1]]
        outputs = layer.outputs
        layer = matches[layers_id[96]]
        layer.inputs.pop("scale_factor")
        layer.inputs["size"] = size
        layer.outputs = outputs
        layers.append(layer)
        return layers



#   Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from x2paddle.optimizer.pytorch_optimizer.fusion import *
from x2paddle.optimizer.pytorch_optimizer.pass_manager import PassManager


class GraphOptimizer(object):
    def __init__(self):
        self.passes = [
            "constant_fuse_pass", "batchnorm2d_fuse_pass",
            "interpolate_bilinear_fuse_pass", "fc_fuse_pass",
            "adaptive_pool2d_fuse_pass", "reshape_fuse_pass",
            "dropout_fuse_pass"
        ]

    def optimize(self, graph):
        for pass_name in self.passes:
            pass_ = PassManager.lookup(pass_name)()
            pass_.apply(graph)
            print("{} done!".format(pass_name))
        return graph



import copy


class BiasOpt:
    def __init__(self):
        self.conv_layers = [
            'fluid.layers.conv2d', 'fluid.layers.conv2d_transpose'
        ]
        self.act_layers = [
            'fluid.layers.relu', 'fluid.layers.relu6', 'fluid.layers.sigmoid',
            'fluid.layers.exp', 'fluid.layers.tanh', 'fluid.layers.softplus',
            'fluid.layers.leaky_relu'
        ]

    def run(self, graph):
        print("Optimize: BiasOpt...")
        layers = copy.deepcopy(graph.layers)
        for layer_id, layer in layers.items():
            if layer.kernel in self.conv_layers or layer.kernel == "fluid.layers.transpose":
                if len(graph.edges_out.get(layer_id, [])) > 1:
                    continue
                if layer.outputs[0] in graph.outputs:
                    continue

                out_layer_id = graph.edges_out[layer_id][0]
                if graph.layers[
                        out_layer_id].kernel != "fluid.layers.elementwise_add":
                    continue
                if graph.layers[out_layer_id].attrs.get('axis', -1) != -1:
                    continue

                in_layer_id = graph.edges_in[out_layer_id]
                bias_layer_id = in_layer_id[1 - in_layer_id.index(layer_id)]
                if graph.layers[
                        bias_layer_id].kernel != "fluid.layers.create_parameter":
                    continue

                bias_layer = graph.layers[bias_layer_id]
                if len(bias_layer.attrs['shape']) != 1:
                    continue
                if len(graph.edges_out[bias_layer_id]) != 1:
                    continue

                if layer.kernel == "fluid.layers.transpose":
                    if layer.attrs['perm'] != [0, 2, 3, 1]:
                        continue
                    in_layer_id = graph.edges_in[layer_id][0]
                    if graph.layers[in_layer_id].kernel not in self.conv_layers:
                        continue
                    if graph.layers[in_layer_id].attrs['bias_attr'] != False:
                        continue
                    if graph.layers[in_layer_id].outputs[0] in graph.outputs:
                        continue
                    if len(graph.edges_out[in_layer_id]) != 1:
                        continue
                    graph.layers[in_layer_id].attrs[
                        'bias_attr'] = bias_layer.attrs['name']
                else:
                    graph.layers[layer_id].attrs[
                        'bias_attr'] = bias_layer.attrs['name']
                bias_add_outs = graph.edges_out.get(out_layer_id, [])
                bias_add_output = graph.layers[out_layer_id].outputs[0]
                graph.del_layer(bias_layer_id)
                graph.del_layer(out_layer_id)

                for out in bias_add_outs:
                    for k, v in graph.layers[out].inputs.items():
                        if v == layer.outputs[0]:
                            graph.layers[out].inputs[k] = bias_add_output
                graph.layers[layer_id].outputs[0] = bias_add_output

                if layer.kernel == "fluid.layers.transpose":
                    in_layer_id = graph.edges_in[layer_id][0]
                    graph.layers[in_layer_id].outputs[0] = bias_add_output
                    graph.layers[layer_id].inputs['x'] = bias_add_output



import copy
import sys


class TransposeOpt:
    def __init__(self):
        self.image_layers = [
            'fluid.layers.conv2d', 'fluid.layers.batch_norm',
            'fluid.layers.conv2d_transpose', 'fluid.layers.resize_nearest',
            'fluid.layers.resize_bilinear', 'fluid.layers.pool2d',
            'fluid.layers.pad2d'
        ]
        self.direct_layers = [
            'fluid.layers.relu', 'fluid.layers.relu6', 'fluid.layers.abs',
            'fluid.layers.sigmoid', 'fluid.layers.exp', 'fluid.layers.rsqrt',
            'fluid.layers.swish_f32', 'fluid.layers.tanh',
            'fluid.layers.softplus', 'fluid.layers.leaky_relu',
            'fluid.layers.floor', 'fluid.layers.erf', 'fluid.layers.swish'
        ]
        self.elementwise_layers = [
            'fluid.layers.elementwise_add', 'fluid.layers.elementwise_sub',
            'fluid.layers.elementwise_mul', 'fluid.layers.elementwise_div'
        ]
        #        self.reduce_layers = []
        self.reduce_layers = [
            'fluid.layers.reduce_mean', 'fluid.layers.reduce_all',
            'fluid.layers.reduce_max', 'fluid.layers.reduce_any',
            'fluid.layers.reduce_sum', 'fluid.layers.reduce_prod'
        ]

    def get_transpose_num(self, graph):
        count = 0
        for layer_id, layer in graph.layers.items():
            if layer.kernel == "fluid.layers.transpose":
                count += 1
        return count

    def run(self, graph):
        print("Optimize: TransposeOpt...")
        total_layer_num = len(graph.layers)
        scanned_layers = set()
        optimized_transpose_layers = list()
        optimized_reduce_layers = list()
        optimized_concat_layers = list()
        optimized_elementwise_layers = list()

        def strip_transpose(_graph):
            layers = copy.deepcopy(_graph.layers)
            for layer_id, layer in layers.items():
                if layer_id in scanned_layers:
                    continue
                scanned_layers.add(layer_id)
                percent = round(len(scanned_layers) / total_layer_num * 100, 2)
                sys.stderr.write("\rOptimize Transpose Layers...{}%".format(
                    percent))

                if layer.kernel != "fluid.layers.transpose":
                    continue
                if layer.attrs["perm"] != [0, 2, 3, 1]:
                    continue
                transpose_layers = list()
                propagate_layers = list()
                reduce_layers = list()
                concat_layers = list()
                # 此elementwise_layers专用于存储shape(4) + shape(1)的形式layer
                elementwise_layers = list()
                can_be_optimized = True
                for out in _graph.edges_out.get(layer_id, []):
                    if _graph.layers[out].kernel == "fluid.layers.transpose":
                        if _graph.layers[out].attrs["perm"] != [0, 3, 1, 2]:
                            can_be_optimized = False
                            break
                        transpose_layers.append(out)
                    elif _graph.layers[out].kernel in self.elementwise_layers:
                        propagate_layers.append(out)
                    elif _graph.layers[out].kernel in self.direct_layers:
                        if _graph.layers[out].outputs[0] in _graph.outputs:
                            can_be_optimized = False
                            break
                        propagate_layers.append(out)
                    elif _graph.layers[out].kernel in self.reduce_layers:
                        if _graph.layers[out].outputs[0] in _graph.outputs:
                            can_be_optimized = False
                            break
                        if not _graph.layers[out].attrs.get('keep_dim', False):
                            can_be_optimized = False
                            break
                        propagate_layers.append(out)
                        reduce_layers.append(out)
                    elif _graph.layers[out].kernel == "fluid.layers.concat":
                        if _graph.layers[out].outputs[0] in _graph.outputs:
                            can_be_optimized = False
                            break
                        propagate_layers.append(out)
                        concat_layers.append(out)
                    else:
                        can_be_optimized = False
                        break

                visited_layers = set()
                while len(propagate_layers) > 0 and can_be_optimized:
                    current_id = propagate_layers.pop(0)
                    visited_layers.add(current_id)
                    for out in _graph.edges_out.get(current_id, []):
                        if _graph.layers[
                                out].kernel == "fluid.layers.transpose":
                            if _graph.layers[out].attrs["perm"] != [0, 3, 1, 2]:
                                can_be_optimized = False
                                break
                            transpose_layers.append(out)
                        elif _graph.layers[
                                out].kernel in self.elementwise_layers:
                            if _graph.layers[out].outputs[0] in _graph.outputs:
                                can_be_optimized = False
                                break
                            if out not in visited_layers:
                                propagate_layers.append(out)
                        elif _graph.layers[out].kernel in self.direct_layers:
                            if _graph.layers[out].outputs[0] in _graph.outputs:
                                can_be_optimized = False
                                break
                            if out not in visited_layers:
                                propagate_layers.append(out)
                        elif _graph.layers[out].kernel in self.reduce_layers:
                            if _graph.layers[out].outputs[0] in _graph.outputs:
                                can_be_optimized = False
                                break
                            if not _graph.layers[out].attrs.get('keep_dim',
                                                                False):
                                can_be_optimized = False
                                break
                            if out not in visited_layers:
                                propagate_layers.append(out)
                                reduce_layers.append(out)
                        elif _graph.layers[out].kernel == "fluid.layers.concat":
                            if _graph.layers[out].outputs[0] in _graph.outputs:
                                can_be_optimized = False
                                break
                            if out not in visited_layers:
                                propagate_layers.append(out)
                                concat_layers.append(out)
                        else:
                            can_be_optimized = False
                            break
                    for ipt in _graph.edges_in.get(current_id, []):
                        if _graph.layers[
                                current_id].kernel in self.elementwise_layers:
                            try:
                                x_shape = _graph.layers[
                                    current_id].input_shapes['x']
                                y_shape = _graph.layers[
                                    current_id].input_shapes['y']
                                if _graph.layers[ipt].outputs[
                                        0] == _graph.layers[current_id].inputs[
                                            'x']:
                                    if len(x_shape) <= 1:
                                        elementwise_layers.append(current_id)
                                        continue
                                elif _graph.layers[ipt].outputs[
                                        0] == _graph.layers[current_id].inputs[
                                            'y']:
                                    if len(y_shape) <= 1:
                                        elementwise_layers.append(current_id)
                                        continue
                                else:
                                    raise Exception(
                                        "Unexcepted situation happend while optimizing transpose"
                                    )
                            except Exception as e:
                                can_be_optimized = False
                                break
                        if _graph.layers[
                                ipt].kernel == "fluid.layers.transpose":
                            if _graph.layers[ipt].attrs["perm"] != [0, 2, 3, 1]:
                                can_be_optimized = False
                                break
                            if ipt not in visited_layers:
                                transpose_layers.append(ipt)
                        elif _graph.layers[
                                ipt].kernel in self.elementwise_layers:
                            if _graph.layers[ipt].outputs[0] in _graph.outputs:
                                can_be_optimized = False
                                break
                            if ipt not in visited_layers:
                                propagate_layers.append(ipt)
                        elif _graph.layers[ipt].kernel in self.direct_layers:
                            if _graph.layers[ipt].outputs[0] in _graph.outputs:
                                can_be_optimized = False
                                break
                            if ipt not in visited_layers:
                                propagate_layers.append(ipt)
                        elif _graph.layers[ipt].kernel in self.reduce_layers:
                            if _graph.layers[ipt].outputs[0] in _graph.outputs:
                                can_be_optimized = False
                                break
                            if not _graph.layers[ipt].attrs.get('keep_dim',
                                                                False):
                                can_be_optimized = False
                                break
                            if ipt not in visited_layers:
                                propagate_layers.append(ipt)
                                reduce_layers.append(ipt)
                        elif _graph.layers[ipt].kernel == "fluid.layers.concat":
                            if _graph.layers[ipt].outputs[0] in _graph.outputs:
                                can_be_optimized = False
                                break
                            if ipt not in visited_layers:
                                propagate_layers.append(ipt)
                                concat_layers.append(ipt)
                        else:
                            can_be_optimized = False
                            break
                    if not can_be_optimized:
                        break
                if not can_be_optimized:
                    continue

                transpose_layers.append(layer_id)
                transpose_layers = list(set(transpose_layers))
                for l in transpose_layers:
                    if graph.layers[l].outputs[0] in graph.outputs:
                        can_be_optimized = False
                        break
                if not can_be_optimized:
                    continue

                for l in transpose_layers:
                    _graph.del_layer(l)

                optimized_transpose_layers.extend(transpose_layers)
                optimized_reduce_layers.extend(reduce_layers)
                optimized_concat_layers.extend(concat_layers)
                optimized_elementwise_layers.extend(elementwise_layers)
                return True
            return False

        before_transpose_num = self.get_transpose_num(graph)
        opt_graph = copy.deepcopy(graph)
        total_layer_num = len(opt_graph.layers)

        while strip_transpose(opt_graph):
            pass

        for layer_id in list(set(optimized_transpose_layers)):
            graph.del_layer(layer_id)
        for layer_id in list(set(optimized_reduce_layers)):
            dim = graph.layers[layer_id].attrs.get('dim', None)
            if dim is not None:
                for i in range(len(dim)):
                    dim[i] = [0, 2, 3, 1][dim[i]]
                graph.layers[layer_id].attrs['dim'] = dim
        for layer_id in list(set(optimized_concat_layers)):
            axis = graph.layers[layer_id].attrs.get('axis', 0)
            graph.layers[layer_id].attrs['axis'] = [0, 2, 3, 1][axis]
        for layer_id in list(set(optimized_elementwise_layers)):
            axis = graph.layers[layer_id].attrs.get('axis', -1)
            graph.layers[layer_id].attrs['axis'] = [0, 2, 3, 1][axis]

        current_transpose_num = self.get_transpose_num(graph)
        print(
            "\nTranspose layers optimized, before: transpose_num={}, after: transpose_num={}".
            format(before_transpose_num, current_transpose_num))



#   Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from x2paddle.optimizer.pytorch_optimizer.pattern_matcher import FuseBase
from x2paddle.core.program import PaddleGraph, PaddleLayer
from x2paddle.core.util import *


class AdaptivePool2dFuser(FuseBase):
    def __init__(self):
        super(AdaptivePool2dFuser, self).__init__(graph_type="dygraph")

    def build_pattern(self):
        """ 描述需要替换的adaptive pool2d图结构。
        adaptive pool2d层模式python实现代码示例:
            x68 = fluid.layers.shape(input=x60)
            x69 = len(x68)
            x70 = x69 <= 2
            if x70 :
                raise RaiseException('Exception')
            x73 = []
            x74 = x68[-2: 2147483647: 1]
            x75 = len(x74)
            x76 = [2, x75]
            x77 = min(x76)
            for _x79 in range(x77):
                x80 = [6, 6][_x79]
                x73.append(x80)
            x81 = fluid.layers.adaptive_pool2d(input=x60, pool_size=x73, pool_type='avg')
        """

        def gen_name(id):
            return "x" + str(id)

        self.pattern.add_layer(
            "fluid.layers.shape",
            inputs={'input': "pool-input-0"},
            outputs=[gen_name(1)])
        self.pattern.add_layer(
            "prim.len", inputs={"input": gen_name(1)}, outputs=[gen_name(6)])
        self.pattern.add_layer(
            "prim.le", inputs={"x": gen_name(6)}, outputs=[gen_name(8)], y=2)
        self.pattern.add_layer("prim.if", {'input': gen_name(8)}, [gen_name(9)])
        if_layer = self.pattern.layers[list(self.pattern.layers.keys())[-1]]
        pattern_block0 = PaddleGraph(if_layer, graph_type="dygraph")
        pattern_block0.add_layer(
            "prim.exception",
            inputs={},
            outputs=[gen_name(9)],
            input="Exception")
        if_layer.add_block(pattern_block0)
        pattern_block1 = PaddleGraph(if_layer, graph_type="dygraph")
        if_layer.add_block(pattern_block1)
        self.pattern.add_layer("prim.list", inputs={}, outputs=[gen_name(10)])
        self.pattern.add_layer(
            "prim.slice",
            inputs={"input": gen_name(1), },
            outputs=[gen_name(12)],
            start=-1,
            end=100,
            step=1)
        self.pattern.add_layer(
            "prim.len", inputs={"input": gen_name(12)}, outputs=[gen_name(14)])
        self.pattern.add_layer(
            "prim.list",
            inputs={"input1": gen_name(14)},
            outputs=[gen_name(15)],
            input0=2)
        self.pattern.add_layer(
            "prim.min", inputs={"input": gen_name(15)}, outputs=[gen_name(16)])
        self.pattern.add_layer("prim.loop", {'input': gen_name(16)},
                               [gen_name(17), gen_name(18)])
        loop_layer = self.pattern.layers[list(self.pattern.layers.keys())[-1]]
        pattern_block = PaddleGraph(loop_layer, graph_type="dygraph")
        pattern_block.add_layer(
            "prim.getitem",
            inputs={"index": gen_name(18)},
            outputs=[gen_name(19)],
            list=[6, 6])
        pattern_block.add_layer(
            "prim.append",
            inputs={"list": gen_name(10),
                    "index": gen_name(19)},
            outputs=[gen_name(20)])
        loop_layer.inputs["input-0"] = gen_name(10)
        loop_layer.add_block(pattern_block)
        pool_attrs = {'pool_type': string("avg")}
        self.pattern.add_layer(
            "fluid.layers.adaptive_pool2d",
            inputs={'input': "pool-input-0",
                    "pool_size": gen_name(10)},
            outputs=[gen_name(21)],
            **pool_attrs)
        self.pattern.build(inputs={"input-0": "pool-input-0", })

    def insert_new_layer(self, graph, parameters, matches):
        parameters = graph.parameters
        new_layer = self.gen_new_layer(parameters, matches)
        new_layer_id = list(matches.keys())[0]
        graph.layers[new_layer_id] = new_layer
        matches.pop(new_layer_id)

    def gen_new_layer(self, parameters, matches):
        layers_id = list(matches.keys())
        layer = matches[layers_id[11]]
        pool_size = layer.attrs["list"]
        layer = matches[layers_id[0]]
        input_name = layer.inputs["input"]
        layer = matches[layers_id[-1]]
        output_name = layer.outputs[0]
        pool_type = layer.attrs["pool_type"]
        attrs = dict()
        attrs["pool_size"] = pool_size
        attrs["pool_type"] = pool_type
        new_layer = PaddleLayer(
            layers_id[0],
            "fluid.layers.adaptive_pool2d",
            inputs={"input": input_name},
            outputs=[output_name],
            **attrs)
        return new_layer



from .register import register
from x2paddle.core.util import *


def shufflechannel_shape(input_shape):
    return input_shape


def shufflechannel_layer(inputs, group=None, input_shape=None, name=None):
    input = inputs[0]
    out = fluid.layers.shuffle_channel(x=input, group=group)
    return out


def shufflechannel_weights(name, data=None):
    weights_name = []
    return weights_name


register(
    kind='ShuffleChannel',
    shape=shufflechannel_shape,
    layer=shufflechannel_layer,
    weights=shufflechannel_weights)



# Copyright (c) 2019  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from x2paddle.decoder.onnx_decoder import ONNXGraph, ONNXGraphNode, ONNXGraphDataNode
from x2paddle.core.graph import GraphNode
from x2paddle.core.fluid_code import Layer
from x2paddle.core.fluid_code import FluidCode
from x2paddle.core.util import string
from x2paddle.op_mapper.onnx2paddle.opset9.custom_layer import *
from functools import reduce
import numpy as np
import onnx
import onnx.numpy_helper as numpy_helper
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
import logging as _logging
from collections import OrderedDict
import math
import os
import shutil

_logger = _logging.getLogger(__name__)


def _const_weight_or_none(node, necessary=False):
    if 'Constant' in node.layer_type:
        return node.value
    if isinstance(node, ONNXGraphDataNode):
        return node.weight
    if necessary:
        assert '{} should be an initializer or Constant operator.'.format(
            node.layer_name)
    return None


def _is_static_shape(shape):
    negtive_dims = 0
    error_dims = 0
    for dim in shape:
        if dim < 0:
            negtive_dims += 1
        if dim < -1:
            error_dims += 1
    if negtive_dims > 1:
        return False
    if error_dims > 0:
        return False
    return True


def _get_same_padding(in_size, kernel_size, stride):
    new_size = int(math.ceil(in_size * 1.0 / stride))
    pad_size = (new_size - 1) * stride + kernel_size - in_size
    pad0 = int(pad_size / 2)
    pad1 = pad_size - pad0
    return [pad0, pad1]


def print_mapping_info(func):
    def run_mapping(*args, **kwargs):
        node = args[1]
        try:
            res = func(*args, **kwargs)
        except:
            print("convert failed node:{}, op_type is {}".format(
                node.layer_name[9:], node.layer_type))
            raise
        else:
            #print("convert successfully node:{}, op_type is {}".format(
            #    node.layer_name[9:], node.layer_type))
            return res

    return run_mapping


class OpSet9():
    elementwise_ops = {
        'Add': 'elementwise_add',
        'Div': 'elementwise_div',
        'Sub': 'elementwise_sub',
        'Mul': 'elementwise_mul',
        'Pow': 'elementwise_pow',
    }

    default_op_mapping_field_values = OrderedDict()
    default_op_mapping_field_values['FLUID_OP'] = ''
    default_op_mapping_field_values['FLUID_INPUT_ARGS'] = None
    default_op_mapping_field_values['FLUID_OUTPUT_ARGS'] = None
    default_op_mapping_field_values['ATTR_MAPPING'] = dict()
    default_op_mapping_field_values['DEFAULTS'] = dict()
    default_op_mapping_field_values['INPUT_PERM'] = None
    default_op_mapping_field_values['OUTPUT_PERM'] = None
    default_op_mapping_field_values['FILL_NAME_FIELD'] = True

    default_op_mapping = {
        'Shape': ['shape', ['X'], ['Out']],
        'Erf': ['erf', ['X'], ['Out']],
        'Ceil': ['ceil', ['X'], ['Out']],
        'ReduceMean': [
            'reduce_mean', ['X'], ['Out'], dict(
                axes='dim', keepdims='keep_dim'), dict(keep_dim=1)
        ],
        'ReduceSum': [
            'reduce_sum', ['X'], ['Out'], dict(
                axes='dim', keepdims='keep_dim'), dict(keep_dim=1)
        ],
        'ReduceMin': [
            'reduce_min', ['X'], ['Out'], dict(
                axes='dim', keepdims='keep_dim'), dict(keep_dim=1)
        ],
        'ReduceMax': [
            'reduce_max', ['X'], ['Out'], dict(
                axes='dim', keepdims='keep_dim'), dict(keep_dim=1)
        ],
        #active function
        'Relu': ['relu', ['X'], ['Out']],
        'LeakyRelu': ['leaky_relu', ['X'], ['Out'], dict(), dict(alpha=.01)],
        'Elu': ['elu', ['X'], ['Out'], dict(), dict(alpha=1.)],
        'ThresholdedRelu': [
            'thresholded_relu', ['X'], ['Out'], dict(alpha='threshold'),
            dict(alpha=1.)
        ],
        'Tanh': ['tanh', ['X'], ['Out']],
        'Sigmoid': ['sigmoid', ['X'], ['Out']],
        'HardSigmoid': [
            'hard_sigmoid', ['X'], ['Out'], dict(
                alpha='slope', beta='offset'), dict(
                    slope=.2, offset=.5)
        ],
        'Softsign': ['softsign', ['X'], ['Out']],
        'Softplus': ['softplus', ['X'], ['Out']],
        'Exp': ['exp', ['X'], ['Out']],
        'Softmax': ['softmax', ['X'], ['Out'], dict(), dict(axis=1)],
        'Sqrt': ['sqrt', ['X'], ['Out']],
        'Floor': ['floor', ['X'], ['Out']],
        'Abs': ['abs', ['X'], ['Out']],
    }

    default_ioa_constraint = {}

    def __init__(self, decoder):
        super(OpSet9, self).__init__()
        self.graph = decoder.graph
        self.input_shapes = []
        self.weights = dict()
        self.omit_nodes = list()
        self.used_custom_layers = dict()

    @print_mapping_info
    def directly_map(self, node, name='', *args, **kwargs):
        inputs = node.layer.input
        outputs = node.layer.output
        op_type = node.layer_type
        attrs = node.attr_map
        info = self.default_op_mapping[op_type]
        info.extend(
            list(self.default_op_mapping_field_values.values())[len(info):])
        (
            fluid_op,
            fluid_input_args,
            fluid_output_args,
            attr_mapping,
            default_attrs,
            input_perm,
            output_perm,
            fill_name_field, ) = info

        if fluid_op in self.default_ioa_constraint:
            for predicate, message in self.default_ioa_constraint[fluid_op]:
                assert predicate(inputs, outputs, attrs), message

        mapped_attrs = {
            attr_mapping.get(key, key): value
            for key, value in attrs.items()
        }
        if '' in mapped_attrs:
            mapped_attrs.pop('')
        if '_' in mapped_attrs:
            mapped_attrs.pop('_')
        fluid_attrs = default_attrs.copy()
        fluid_attrs.update(mapped_attrs)
        inputs = inputs if input_perm is None else list(
            map(lambda i: inputs[i], input_perm))
        val_inps = []
        for idx, ipt in enumerate(inputs):
            val_inps.append(self.graph.get_input_node(node, idx=idx, copy=True))

        val_outs = outputs if output_perm is None else list(
            map(lambda i: outputs[i], output_perm))
        attr = fluid_attrs
        assert len(val_inps) == 1, 'directly_map error with multi inputs'
        if fluid_op not in ['shape', 'erf']:
            attr['name'] = string(node.layer_name)
        node.fluid_code.add_layer(
            fluid_op, inputs=val_inps[0], output=val_outs[0], param_attr=attr)
        if fluid_op in ['shape']:
            node.fluid_code.add_layer(
                'cast',
                inputs=val_outs[0],
                output=val_outs[0],
                param_attr={'dtype': string('int64')})

    @print_mapping_info
    def deal_custom_layer(self, node):
        op = node.layer_type
        custom_code, func = make_custom_layer(node)
        child_func_code, child_func = make_custom_child_func(node)
        params = get_params(node.layer, node.layer_type)
        arg_names, kwargs = set_args(func, params)
        kwargs['name'] = string(node.layer_name)
        node.fluid_code.add_layer(
            func.__code__.co_name,
            inputs=node.inputs,
            output=node,
            param_attr=kwargs,
            is_custom_layer=True)
        if op not in self.used_custom_layers:
            self.used_custom_layers[op] = custom_code
            if op + '_child_func' not in self.used_custom_layers:
                if child_func_code is not None:
                    self.used_custom_layers[op +
                                            '_child_func'] = child_func_code

    @print_mapping_info
    def elementwise_map(self, node):
        assert node.layer_type in self.elementwise_ops
        op_type = self.elementwise_ops[node.layer_type]

        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_y = self.graph.get_input_node(node, idx=1, copy=True)
        inputs = {'x': val_x, 'y': val_y}
        node.fluid_code.add_layer(
            op_type, inputs=inputs, output=node, param_attr=None)

    @print_mapping_info
    def place_holder(self, node):
        self.input_shapes.append(node.out_shapes[0])

        shape = node.out_shapes[0]
        for i, dim_shape in enumerate(shape):
            if dim_shape == 0 and i == 0:
                shape[i] = 1
            if dim_shape == 0 and i != 0:
                assert 'shape of input is not assigned'
        attr = {
            "dtype": string(node.dtype),
            "shape": shape,
            "name": string(node.layer_name),
            "append_batch_size": 'False'
        }

        node.fluid_code.add_layer(
            "data", inputs=None, output=node, param_attr=attr)

    @print_mapping_info
    def create_parameter(self, node, parameter=None):
        if parameter is not None:
            node = parameter
        dtype = node.dtype
        shape = node.out_shapes[0]
        if len(node.weight.shape) == 0:
            shape = [1]
        self.weights[node.layer_name] = node.weight
        attr = {
            'dtype': string(dtype),
            'shape': shape,
            'name': string(node.layer_name),
            'default_initializer': 'Constant(0.0)'
        }
        if dtype == 'bool':
            attr['dtype'] = string('int64')
            node.fluid_code.add_layer(
                "create_parameter", inputs=None, output=node, param_attr=attr)
            node.fluid_code.add_layer(
                "cast",
                inputs=node,
                output=node,
                param_attr={'dtype': string('bool')})
        elif dtype == 'uint8':
            attr['dtype'] = string('float32')
            node.fluid_code.add_layer(
                "create_parameter", inputs=None, output=node, param_attr=attr)
        else:
            node.fluid_code.add_layer(
                "create_parameter", inputs=None, output=node, param_attr=attr)

    def _pad_if_asymmetric(self, node, pads, val_name):  # pads: SSEE
        assert len(pads) & 1 == 0
        symmetric = True
        ndims = len(pads) // 2
        for idx_dim in range(ndims):
            if pads[idx_dim] != pads[ndims + idx_dim]:
                symmetric = False
                break
        if symmetric:
            return pads[:ndims], val_name
        val_padded = self.Pad(node, op_independent=False)
        return [0] * ndims, val_padded

    def _interpolate(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        inputs = {'input': val_x}
        if node.layer_type == 'Resize':
            if len(node.layer.input) == 2:
                # opset 10
                val_scales = self.graph.get_input_node(node, idx=1, copy=True)
                inputs['scale'] = val_scales
            elif len(node.layer.input) == 3:
                # opset 11
                val_scales = self.graph.get_input_node(node, idx=2, copy=True)
                inputs['scale'] = val_scales
            elif len(node.layer.input) == 4:
                # opset 11
                val_sizes = self.graph.get_input_node(node, idx=3, copy=True)
                var_nc, var_hw = val_sizes.layer_name + '_nc', val_sizes.layer_name + '_hw'
                node.fluid_code.add_layer(
                    'split',
                    inputs=val_sizes,
                    output=var_nc + ',' + var_hw,
                    param_attr={
                        'dim': 0,
                        'num_or_sections': [2, 2],
                    })
                node.fluid_code.add_layer(
                    "cast",
                    inputs=var_hw,
                    output=var_hw,
                    param_attr={'dtype': string('int32')})
                inputs['out_shape'] = var_hw
        elif node.layer_type == 'Upsample':
            val_scales = self.graph.get_input_node(node, idx=1, copy=True)
            inputs['scale'] = val_scales

        attr = {'name': string(node.layer_name)}
        mode = node.get_attr('mode', 'nearest')
        fluid_op = 'resize_{}'.format(mode)
        if 'linear' in mode:
            print(
                'Warnning: paddle not support op:resize wiht mode: linear, we use bilinear replace linear'
            )
            fluid_op = 'resize_bilinear'
        attr['align_corners'] = False
        node.fluid_code.add_layer(
            fluid_op, inputs=inputs, output=node, param_attr=attr)

    @print_mapping_info
    def RoiAlign(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_rois = self.graph.get_input_node(node, idx=1, copy=True)

        pooled_height = node.get_attr('output_height')
        pooled_width = node.get_attr('output_width')
        spatial_scale = node.get_attr('spatial_scale')
        sampling_ratio = node.get_attr('sampling_ratio')
        attr = {
            'pooled_height': pooled_height,
            'pooled_width': pooled_width,
            'spatial_scale': spatial_scale,
            'sampling_ratio': sampling_ratio,
        }
        node.fluid_code.add_layer(
            'roi_align',
            inputs={'input': val_x,
                    'rois': val_rois},
            output=node,
            param_attr=attr)

    @print_mapping_info
    def MaxRoiPool(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_rois = self.graph.get_input_node(node, idx=1, copy=True)

        spatial_scale = node.get_attr('spatial_scale')
        pooled_height, pooled_width = node.get_attr('pooled_shape')
        attr = {
            'pooled_height': pooled_height,
            'pooled_width': pooled_width,
            'spatial_scale': spatial_scale,
        }
        node.fluid_code.add_layer(
            'roi_pool',
            inputs={'input': val_x,
                    'rois': val_rois},
            output=node,
            param_attr=attr)

    @print_mapping_info
    def Pad(self, node, op_independent=True):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        pads = node.get_attr('pads')
        mode = node.get_attr('mode', 'constant')
        value = node.get_attr('value', 0.)
        data_shape = val_x.out_shapes[0]
        output_shape = node.out_shapes[0]
        assume_pad2d = False
        attr = {}
        paddings = []
        if len(pads) == 4:
            assume_pad2d |= mode != 'constant'
            if data_shape:
                assume_pad2d |= data_shape and len(data_shape) == 4  # NCHW
            if output_shape:
                assume_pad2d |= output_shape and len(output_shape) == 4  # NCHW
        if assume_pad2d:
            fluid_op = 'pad2d'
            attr['data_format'] = string('NCHW')
            attr['mode'] = string(mode)
        else:
            attr = {'pad_value': value}
            fluid_op = 'pad'
        if len(pads) == 4:
            paddings = np.array(pads).reshape(
                (-1, 2)).transpose().flatten().tolist()  # SSEE -> SESE
        elif len(pads) == 8:
            paddings = np.array(pads).reshape(
                (-1, 4)).transpose().flatten().tolist()  # SSEE -> SESE
            if sum(paddings[:4]) == 0:
                fluid_op = 'pad2d'
                paddings = paddings[4:]
                attr['mode'] = string(mode)
        attr['paddings'] = paddings
        if op_independent:
            attr['name'] = string(node.layer_name)
            node.fluid_code.add_layer(
                fluid_op, inputs=val_x, output=node, param_attr=attr)
        else:
            attr['name'] = string(node.layer_name + '_paded')
            node.fluid_code.add_layer(
                fluid_op,
                inputs=val_x,
                output=node.layer_name + '_paded',
                param_attr=attr)
            return node.layer_name + '_paded'

    @print_mapping_info
    def Unsqueeze(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        axes = node.get_attr('axes')
        attr = {'axes': axes, 'name': string(node.layer_name)}
        if len(val_x.out_shapes[0]) == 0:
            if node.layer_name:
                node.fluid_code.add_layer(
                    'reshape',
                    inputs=val_x,
                    output=node,
                    param_attr={'shape': [1]})
        else:
            if str(val_x.dtype) == 'bool':
                val_x_cast = val_x.layer_name + '_cast'
                node.fluid_code.add_layer(
                    'cast',
                    inputs=val_x,
                    output=val_x_cast,
                    param_attr={'dtype': string('int64')})
                node.fluid_code.add_layer(
                    'unsqueeze',
                    inputs=val_x_cast,
                    output=node,
                    param_attr=attr)
            else:
                node.fluid_code.add_layer(
                    'unsqueeze', inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def Shrink(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        bias = node.get_attr('bias')
        lambd = node.get_attr('lambd')
        assert bias == 0.0, 'not support bias!=0'
        attr = {'threshold': lambd, 'name': node.layer_name}
        node.fluid_code.add_layer(
            'hard_shrink', inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def Constant(self, node):
        val_output = self.graph.get_node(node.layer.output[0], copy=True)

        value = node.get_attr('value')
        dtype = np.dtype(value.dtype)
        output_dtype = val_output.dtype
        if output_dtype:
            assert dtype == output_dtype, 'tensor dtype unmatches storage dtype'

        shape = node.get_attr('shape', None)

        if shape is None:
            shape = val_output.out_shapes[0]
        if shape is None:
            shape = list(value.shape)
            _logger.warning('in (Constant -> %s): '
                            'attribute "shape" of %s not inferred, '
                            'using value as 1-D tensor may lead to fails',
                            val_output.layer_name, val_output.layer_name)
        if len(value) == 1:
            value = value.tolist()
            shape = [1]
            value = value[0]
            if dtype.name == 'int64':
                dtype = 'int32'
            attr = {'shape': shape, 'dtype': string(dtype), 'value': value}
            node.fluid_code.add_layer(
                'fill_constant', inputs=None, output=node, param_attr=attr)
        else:
            if dtype.name == 'uint8':
                dtype = 'int64'
            value = np.reshape(value, shape)
            self.weights[node.layer_name] = value
            attr = {
                'dtype': string(dtype),
                'shape': shape,
                'name': string(node.layer_name),
                'default_initializer': 'Constant(0.0)'
            }
            node.fluid_code.add_layer(
                "create_parameter", inputs=None, output=node, param_attr=attr)

    @print_mapping_info
    def Resize(self, node):
        self._interpolate(node)

    @print_mapping_info
    def Upsample(self, node):
        self._interpolate(node)

    @print_mapping_info
    def InstanceNormalization(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_scale = self.graph.get_input_node(node, idx=1, copy=True)
        val_b = self.graph.get_input_node(node, idx=2, copy=True)
        epsilon = node.get_attr('epsilon', 1e-5)
        attr = {
            'epsilon': epsilon,
            'param_attr': string(val_scale.layer_name),
            'bias_attr': string(val_b.layer_name)
        }
        node.fluid_code.add_layer(
            "instance_norm", inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def Expand(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_shape = self.graph.get_input_node(node, idx=1, copy=True)
        val_x_dtype = val_x.dtype
        name_ones = node.layer_name + '_ones'
        attr_ones = {
            'shape': val_shape.layer_name,
            'dtype': string(val_x_dtype),
            'value': 1
        }
        node.fluid_code.add_layer(
            'fill_constant',
            inputs=None,
            output=name_ones,
            param_attr=attr_ones)
        inputs = {'x': name_ones, 'y': val_x}
        node.fluid_code.add_layer(
            'elementwise_mul',
            inputs=inputs,
            output=node.layer_name,
            param_attr=None)

    @print_mapping_info
    def Gather(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        indices = self.graph.get_input_node(node, idx=1, copy=True)
        indices_shape = indices.out_shapes[0]
        axis = node.get_attr('axis', 0)
        #assert len(
        #    indices_shape) <= 2, "Gather op don't support dim of indice >2 "
        if axis == 0 and len(indices_shape) <= 1:
            if len(val_x.out_shapes[0]) <= 1:
                node.fluid_code.add_layer(
                    'gather',
                    inputs={'input': val_x,
                            'index': indices},
                    output=node,
                    param_attr=None)
            elif len(val_x.out_shapes[0]) > 1:
                if len(indices_shape) == 0:
                    gather_ = node.layer_name + '_1'
                    node.fluid_code.add_layer(
                        'gather',
                        inputs={'input': val_x,
                                'index': indices},
                        output=gather_,
                        param_attr=None)
                    node.fluid_code.add_layer(
                        'squeeze',
                        inputs={'input': gather_,
                                'axes': [0]},
                        output=node,
                        param_attr=None)
                else:
                    node.fluid_code.add_layer(
                        'gather',
                        inputs={'input': val_x,
                                'index': indices},
                        output=node,
                        param_attr=None)
        elif axis > 0 and len(indices_shape) <= 1:
            perm = list(range(len(val_x.out_shapes[0])))
            perm = [axis] + perm[:axis] + perm[axis + 1:]
            attr_trans = {'perm': perm}
            name_trans = val_x.layer_name + '_trans'
            node.fluid_code.add_layer(
                'transpose',
                inputs=val_x,
                output=name_trans,
                param_attr=attr_trans)
            node.fluid_code.add_layer(
                'gather',
                inputs={'input': name_trans,
                        'index': indices},
                output=node,
                param_attr=None)
            node.fluid_code.add_layer(
                'transpose', inputs=node, output=node, param_attr=attr_trans)
            if len(indices_shape) < 1:
                node.fluid_code.add_layer(
                    'squeeze',
                    inputs={'input': node,
                            'axes': [axis]},
                    output=node,
                    param_attr=None)
        elif axis == 0 and len(indices_shape) > 1:
            if val_x.out_shapes[0] is not None and isinstance(
                    val_x, ONNXGraphDataNode):
                indices_cast = indices.layer_name + '_cast'
                node.fluid_code.add_layer(
                    'cast',
                    inputs=indices,
                    output=indices_cast,
                    param_attr={'dtype': string('int64')})
                node.fluid_code.add_layer(
                    'embedding',
                    inputs=indices_cast,
                    output=node,
                    use_fluid=True,
                    param_attr={
                        'param_attr': string(val_x.layer_name),
                        'size': val_x.out_shapes[0]
                    })
            else:
                from functools import reduce
                reshape_shape = reduce(lambda x, y: x * y, indices_shape)
                indices_reshape = indices.layer_name + '_shape'
                node.fluid_code.add_layer(
                    'reshape',
                    inputs=indices,
                    output=indices_reshape,
                    param_attr={'shape': [reshape_shape, ]})

                perm = list(range(len(val_x.out_shapes[0])))
                node.fluid_code.add_layer(
                    'gather',
                    inputs={'input': val_x,
                            'index': indices_reshape},
                    output=node,
                    param_attr=None)
                val_x_shape = val_x.out_shapes[0]
                reshaped_shape = []
                for i in perm:
                    reshaped_shape.append(indices_shape[i])
                for i in val_x_shape[:axis] + val_x_shape[axis + 1:]:
                    reshaped_shape.append(i)
                node.fluid_code.add_layer(
                    'reshape',
                    inputs=node,
                    output=node,
                    param_attr={'shape': reshaped_shape})
        elif axis > 0 and len(indices_shape) > 1:
            from functools import reduce
            reshape_shape = reduce(lambda x, y: x * y, indices_shape)
            indices_reshape = indices.layer_name + '_shape'
            node.fluid_code.add_layer(
                'reshape',
                inputs=indices,
                output=indices_reshape,
                param_attr={'shape': [reshape_shape, ]})

            perm = list(range(len(val_x.out_shapes[0])))
            perm = [axis] + perm[:axis] + perm[axis + 1:]
            attr_trans = {'perm': perm}
            name_trans = val_x.layer_name + '_transpose'
            node.fluid_code.add_layer(
                'transpose',
                inputs=val_x,
                output=name_trans,
                param_attr=attr_trans)
            node.fluid_code.add_layer(
                'gather',
                inputs={'input': name_trans,
                        'index': indices_reshape},
                output=node,
                param_attr=None)
            input_transpose = node.layer_name + '_transpose'
            node.fluid_code.add_layer(
                'transpose',
                inputs=node,
                output=input_transpose,
                param_attr=attr_trans)
            val_x_shape = val_x.out_shapes[0]
            reshaped_shape = []
            for i in perm:
                reshaped_shape.append(indices_shape[i])
            for i in val_x_shape[:axis] + val_x_shape[axis + 1:]:
                reshaped_shape.append(i)
            node.fluid_code.add_layer(
                'reshape',
                inputs=input_transpose,
                output=node,
                param_attr={'shape': reshaped_shape})

    @print_mapping_info
    def ScatterND(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        indices = self.graph.get_input_node(node, idx=1, copy=True)
        updates = self.graph.get_input_node(node, idx=2, copy=True)
        if len(indices.out_shapes[0]) == 1:
            node.fluid_code.add_layer(
                'scatter',
                inputs={'input': val_x,
                        'index': indices,
                        'updates': updates},
                output=node,
                param_attr=None)
        else:
            input_inner_indices = node.layer_name + '_input_inner_indices'
            shape = val_x.out_shapes[0]
            node.fluid_code.add_layer(
                'reshape',
                inputs=indices.layer_name,
                output=indices.layer_name,
                param_attr={'shape': indices.out_shapes[0]})

            zeros_like_val_x = val_x.layer_name + '_zeros'
            node.fluid_code.add_layer(
                'zeros_like',
                inputs=val_x,
                output=zeros_like_val_x,
                param_attr=None)
            node.fluid_code.add_layer(
                'scatter_nd_add',
                inputs={
                    'ref': zeros_like_val_x,
                    'index': indices,
                    'updates': updates
                },
                output=input_inner_indices,
                param_attr=None)
            indices_mask = node.layer_name + '_indices_mask'
            constant_minus_one = node.layer_name + '_constant_minus_one'
            # full_like support create tensor shape like input tensor
            node.fluid_code.add_layer(
                'full_like',
                inputs=updates,
                output=constant_minus_one,
                param_attr={'dtype': string(updates.dtype),
                            'fill_value': -1})
            node.fluid_code.add_layer(
                'scatter_nd_add',
                inputs={
                    'ref': zeros_like_val_x,
                    'index': indices,
                    'updates': constant_minus_one
                },
                output=indices_mask,
                param_attr=None)
            constant_one = node.layer_name + '_constant_1'
            # full_like support create tensor shape like input tensor
            node.fluid_code.add_layer(
                'full_like',
                inputs=val_x,
                output=constant_one,
                param_attr={'dtype': string(val_x.dtype),
                            'fill_value': 1})
            input_out_indices_mask = node.layer_name + '_input_out_indices_mask'
            node.fluid_code.add_layer(
                "elementwise_add",
                inputs={"x": indices_mask,
                        "y": constant_one},
                output=input_out_indices_mask,
                param_attr=None)

            input_out_indices = node.layer_name + '_input_out_indices'
            node.fluid_code.add_layer(
                "elementwise_mul",
                inputs={"x": val_x,
                        "y": input_out_indices_mask},
                output=input_out_indices,
                param_attr=None)

            node.fluid_code.add_layer(
                "elementwise_add",
                inputs={"x": input_inner_indices,
                        "y": input_out_indices},
                output=node,
                param_attr=None)

    @print_mapping_info
    def Range(self, node):
        val_start = self.graph.get_input_node(node, idx=0, copy=True)
        val_limit = self.graph.get_input_node(node, idx=1, copy=True)
        val_delta = self.graph.get_input_node(node, idx=2, copy=True)
        dtype = val_start.dtype
        inputs = {'start': val_start, 'end': val_limit, 'step': val_delta}
        node.fluid_code.add_layer(
            'range',
            inputs=inputs,
            output=node,
            param_attr={'dtype': string(dtype)})

    @print_mapping_info
    def Slice(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        starts, ends, axes, steps = None, None, None, None
        attr = {}
        if len(node.inputs) > 1:
            starts = self.graph.get_input_node(node, idx=1, copy=True)
            ends = self.graph.get_input_node(node, idx=2, copy=True)
            starts_value = _const_weight_or_none(starts)
            ends_value = _const_weight_or_none(ends)

            if len(node.inputs) > 3:
                axes = self.graph.get_input_node(node, idx=3, copy=True)
                axes = _const_weight_or_none(axes, necessary=True)
            if len(node.inputs) > 4:
                steps = self.graph.get_input_node(node, idx=4, copy=True)
                steps = _const_weight_or_none(steps)
            attr = {
                "axes": axes,
                "starts": starts.layer_name,
                "ends": ends.layer_name
            }
            if starts_value is not None and ends_value is not None:
                self.omit_nodes.append(starts.layer_name)
                self.omit_nodes.append(ends.layer_name)
                starts_value = starts_value.copy()
                ends_value = ends_value.copy()
                #for idx in range(len(ends_value)):
                #    if ends_value[idx] > 2**31 - 1:
                #        ends_value[idx] = 2**31 - 1
                #print(val_x.out_shapes)
                for idx in range(len(ends_value)):
                    if starts_value[idx] >= val_x.out_shapes[0][axes[idx]]:
                        starts_value[idx] = val_x.out_shapes[0][axes[idx]] - 1
                        ends_value[idx] = val_x.out_shapes[0][axes[idx]]
                        starts_value[idx] = val_x.out_shapes[0][axes[idx]] - 1
                    elif ends_value[idx] > 2**31 - 1:
                        ends_value[idx] = 2**31 - 1
                attr = {
                    "axes": axes,
                    "starts": starts_value,
                    "ends": ends_value
                }
            else:
                if starts.dtype != 'int32':
                    starts_cast = starts.layer_name + '_cast'
                    node.fluid_code.add_layer(
                        'cast',
                        inputs=starts,
                        output=starts_cast,
                        param_attr={'dtype': string('int32')})
                    attr['starts'] = starts_cast
                if ends.dtype != 'int32':
                    ends_cast = ends.layer_name + '_cast'
                node.fluid_code.add_layer(
                    'cast',
                    inputs=ends,
                    output=ends_cast,
                    param_attr={'dtype': string('int32')})
                attr['ends'] = ends_cast
        else:
            starts = node.get_attr('starts')
            ends = node.get_attr('ends')
            axes = node.get_attr('axes')
            for idx in range(len(ends)):
                if ends[idx] > 2**31 - 1:
                    ends[idx] = 2**31 - 1
            attr = {"axes": axes, "starts": starts, "ends": ends}

        if steps is not None:
            attr['strides'] = steps
            node.fluid_code.add_layer(
                'strided_slice', inputs=val_x, output=node, param_attr=attr)
        else:
            node.fluid_code.add_layer(
                'slice', inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def ConstantOfShape(self, node):
        val_shape = self.graph.get_input_node(node, idx=0, copy=True)
        val_y = self.graph.get_node(node.layer.output[0], copy=True)

        value = node.get_attr('value')
        dtype = value.dtype
        value = value.tolist()
        assert len(value) == 1, ('given value not Scalar, shape of value > 1, '
                                 'this is not supported')
        if len(value) == 1:
            value = value[0]
            attr = {
                'shape': val_shape.layer_name,
                'dtype': string(dtype),
                'value': value
            }
            node.fluid_code.add_layer(
                'fill_constant', inputs=None, output=node, param_attr=attr)

    @print_mapping_info
    def Clip(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_y = self.graph.get_node(node.layer.output[0], copy=True)
        max_value, min_value = None, None
        if len(node.inputs) == 1:
            max_value = node.get_attr('max')
            min_value = node.get_attr('min')
            attr = {
                'max': max_value,
                'min': min_value,
            }
            node.fluid_code.add_layer(
                'clip', inputs=val_x, output=node, param_attr=attr)
        else:
            max_ipt = self.graph.get_input_node(node, idx=1, copy=True)
            min_ipt = self.graph.get_input_node(node, idx=2, copy=True)
            max_value = _const_weight_or_none(max_ipt)
            min_value = _const_weight_or_none(min_ipt)
            self.omit_nodes.append(max_ipt.layer_name)
            self.omit_nodes.append(min_ipt.layer_name)
            if max_value.shape == (1, ):
                max_value = max_value[0]
            if min_value.shape == (1, ):
                min_value = min_value[0]
        if max_value is not None and min_value is not None:
            attr = {'max': max_value, 'min': min_value}
            node.fluid_code.add_layer(
                'clip', inputs=val_x, output=node, param_attr=attr)
        else:
            raise

    @print_mapping_info
    def Split(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_y = self.graph.get_node(node.layer.output[0], copy=True)

        fluid_op = 'split'
        split = node.get_attr('split')
        axis = node.get_attr('axis', 0)
        attr = {
            'num_or_sections': split,
            'dim': axis,
            'name': string(node.layer_name)
        }

        node.fluid_code.add_layer(
            'split', inputs=val_x, output=val_y, param_attr=attr)

    @print_mapping_info
    def Reshape(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_shape = self.graph.get_input_node(node, idx=1, copy=True)
        val_reshaped = self.graph.get_node(node.layer.output[0], copy=True)
        attr = {}
        shape_value = _const_weight_or_none(val_shape)
        shape_dims = len(val_shape.out_shapes[0])

        if shape_value is not None:
            node.fluid_code.add_layer(
                'reshape',
                inputs={'x': val_x},
                output=node,
                param_attr={'shape': shape_value.tolist()})
        elif len(node.out_shapes[0]) > 0 and _is_static_shape(node.out_shapes[
                0]):
            node.fluid_code.add_layer(
                'reshape',
                inputs={'x': val_x,
                        'shape': node.out_shapes[0]},
                output=node,
                param_attr=attr)
        elif val_shape.dtype == 'int64':
            val_shape_cast = val_shape.layer_name + '_cast'
            node.fluid_code.add_layer(
                'cast',
                inputs=val_shape,
                output=val_shape_cast,
                param_attr={'dtype': string('int32')})
            # shape may be [], come form Gather by scalar indices
            if len(val_shape.out_shapes[0]) > 0:
                node.fluid_code.add_layer(
                    'reshape',
                    inputs=val_shape_cast,
                    output=val_shape_cast,
                    param_attr={'shape': val_shape.out_shapes[0]})
            node.fluid_code.add_layer(
                'reshape',
                inputs={'x': val_x,
                        'shape': val_shape_cast},
                output=node,
                param_attr=attr)
        else:
            # shape may be [], come form Gather by scalar indices
            if len(val_shape.out_shapes[0]) > 0:
                node.fluid_code.add_layer(
                    'reshape',
                    inputs=val_shape,
                    output=val_shape,
                    param_attr={'shape': val_shape.out_shapes[0]})
            node.fluid_code.add_layer(
                'reshape',
                inputs={'x': val_x,
                        'shape': val_shape},
                output=node,
                param_attr=attr)

    @print_mapping_info
    def Cast(self, node):
        val_input = self.graph.get_input_node(node, idx=0, copy=True)
        val_output = self.graph.get_node(node.layer.output[0], copy=True)

        dtype = node.get_attr('to')
        if not isinstance(dtype, np.dtype):
            dtype = TENSOR_TYPE_TO_NP_TYPE[dtype]

        output_dtype = val_output.dtype
        if output_dtype:
            assert dtype == output_dtype, 'dtype of to unmatches output'
        attr = {'dtype': string(dtype)}
        node.fluid_code.add_layer(
            'cast', inputs=val_input, output=node, param_attr=attr)

    @print_mapping_info
    def Not(self, node):
        val_input = self.graph.get_input_node(node, idx=0, copy=True)
        node.fluid_code.add_layer('logical_not', inputs=val_input, output=node)

    @print_mapping_info
    def AveragePool(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)

        auto_pad = node.get_attr('auto_pad', 'NOTSET')
        kernel_shape = node.get_attr("kernel_shape")
        poolnd = len(kernel_shape)
        strides = node.get_attr("strides")
        pad_mode = node.get_attr("pads")
        ceil_mode = bool(node.get_attr('ceil_mode', 0))
        pads = node.get_attr('pads', [0] * (poolnd * 2))
        fluid_op = 'pool{}d'.format(poolnd)
        assert 2 <= poolnd <= 3, 'only pool2d and pool3d is supported'

        paddings, val_x = self._pad_if_asymmetric(node, pads, val_x)

        if auto_pad == "SAME_UPPER" or auto_pad == "SAME_LOWER":
            input_shape = val_x.out_shapes[0]
            pad_h = _get_same_padding(input_shape[2], kernel_shape[0],
                                      strides[0])
            pad_w = _get_same_padding(input_shape[3], kernel_shape[1],
                                      strides[1])
            paddings = pad_h + pad_w

        attr = {
            "pool_size": kernel_shape,
            "pool_type": string('avg'),
            "pool_stride": strides,
            "pool_padding": paddings,
            "ceil_mode": ceil_mode,
            "exclusive": 'True',
            "name": string(node.layer_name)
        }

        node.fluid_code.add_layer(
            fluid_op, inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def Concat(self, node):
        inputs = []
        dtypes = set()
        for i in range(len(node.layer.input)):
            ipt = self.graph.get_input_node(node, idx=i, copy=True)
            if isinstance(ipt, str):
                inputs.append(ipt)
            else:
                inputs.append(ipt.layer_name)
                dtypes.add(ipt.dtype)
        if len(dtypes) > 1:
            assert 'Unspported situation happened, please create issue on https://github.com/PaddlePaddle/X2Paddle/issues.'
        axis = node.get_attr('axis')
        attr = {'axis': axis}
        node.fluid_code.add_layer(
            'concat', inputs=inputs, output=node, param_attr=attr)

    @print_mapping_info
    def Flatten(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        axis = node.get_attr('axis', 1)
        attr = {"axis": str(axis), "name": string(node.layer_name)}
        node.fluid_code.add_layer(
            'flatten', inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def Gemm(self, node):
        val_a = self.graph.get_input_node(node, idx=0, copy=True)
        val_b = self.graph.get_input_node(node, idx=1, copy=True)
        val_c = self.graph.get_input_node(node, idx=2, copy=True)

        alpha = node.get_attr('alpha', 1.)  # optional
        beta = node.get_attr('beta', 1.)  # optional
        trans_a = bool(node.get_attr('transA', 0))  # optional
        trans_b = bool(node.get_attr('transB', 0))  # optional
        val_mm = node.layer_name + '_mm'
        matmul_inputs = {"x": val_a, "y": val_b}
        attr_matmul = {
            "transpose_x": trans_a,
            "transpose_y": trans_b,
            "alpha": alpha,
            "name": string(val_mm)
        }
        node.fluid_code.add_layer(
            'matmul',
            inputs=matmul_inputs,
            output=val_mm,
            param_attr=attr_matmul)

        if beta != 0:
            if beta == 1.:
                add_inputs = {"x": val_mm, "y": val_c}
                attr = {"name": string(node.layer_name)}
                node.fluid_code.add_layer(
                    "elementwise_add",
                    inputs=add_inputs,
                    output=node,
                    param_attr=attr)
            else:
                var_beta = node.layer_name + '_beta'
                matmul_beta_inputs = {"x": val_c, "y": var_beta}
                node.fluid_code.add_layer(
                    "Constant",
                    inputs=matmul_beta_inputs,
                    output=var_beta,
                    param_attr={'value': beta})

                add_inputs = {"x": val_mm, "y": var_beta}
                attr = {"name": string(node.layer_name)}
                node.fluid_code.add_layer(
                    "elementwise_add",
                    inputs=add_inputs,
                    output=node,
                    param_attr=attr)

    @print_mapping_info
    def Sum(self, node):
        val_inps = node.layer.input
        inputs = {
            "x": self.graph.get_input_node(
                node, idx=0, copy=True),
            "y": self.graph.get_input_node(
                node, idx=1, copy=True),
        }
        node.fluid_code.add_layer("elementwise_add", inputs=inputs, output=node)

        for idx, ipt in enumerate(val_inps[2:]):
            y = self.graph.get_input_node(node, idx=idx, copy=True)
            inputs = {
                "x": node.layer_name,
                "y": y,
            }
            node.fluid_code.add_layer(
                "elementwise_add", inputs=inputs, output=node)

    @print_mapping_info
    def MatMul(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_y = self.graph.get_input_node(node, idx=1, copy=True)
        x_shape = val_x.out_shapes[0]
        y_shape = val_y.out_shapes[0]
        inputs = {"x": val_x, "y": val_y}
        if y_shape[0] == 1 and x_shape[-1] != 1 and x_shape[0] != 1:
            y_squeeze = val_y.layer_name + '_squeeze'
            node.fluid_code.add_layer(
                "squeeze",
                inputs=val_y,
                output=y_squeeze,
                param_attr={'axes': [0]})
            inputs['y'] = y_squeeze
            node.fluid_code.add_layer(
                "matmul", inputs=inputs, output=node, param_attr=None)
        else:
            node.fluid_code.add_layer(
                "matmul", inputs=inputs, output=node, param_attr=None)

    @print_mapping_info
    def BatchNormalization(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_scale = self.graph.get_input_node(node, idx=1, copy=True)
        val_b = self.graph.get_input_node(node, idx=2, copy=True)
        val_mean = self.graph.get_input_node(node, idx=3, copy=True)
        val_var = self.graph.get_input_node(node, idx=4, copy=True)

        self.omit_nodes.append(val_scale.layer_name)
        self.omit_nodes.append(val_b.layer_name)
        self.omit_nodes.append(val_mean.layer_name)
        self.omit_nodes.append(val_var.layer_name)

        momentum = node.get_attr('momentum', .9)
        epsilon = node.get_attr('epsilon', 1e-5)

        # Attribute: spatial is used in BatchNormalization-1,6,7
        spatial = bool(node.get_attr('spatial'))
        attr = {
            "momentum": momentum,
            "epsilon": epsilon,
            "data_layout": string('NCHW'),
            "is_test": True,
            "param_attr": string(val_scale.layer_name),
            "bias_attr": string(val_b.layer_name),
            "moving_mean_name": string(val_mean.layer_name),
            "moving_variance_name": string(val_var.layer_name),
            "use_global_stats": spatial,
            "name": string(node.layer_name)
        }
        node.fluid_code.add_layer(
            "batch_norm", inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def Transpose(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        perm = node.get_attr('perm')
        attr = {'perm': perm, "name": string(node.layer_name)}
        node.fluid_code.add_layer(
            "transpose", inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def Relu(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        attr = {"name": string(node.layer_name)}
        node.fluid_code.add_layer(
            "relu", inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def PRelu(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_slope = self.graph.get_input_node(node, idx=1, copy=True)

        mode = 'channel'
        shape_slope = val_slope.out_shapes[0]

        if shape_slope == [1]:
            mode = 'all'
        elif len(shape_slope) > 2:
            mode = 'element'

        if mode == 'channel' and len(shape_slope) == 1:
            # paddle params shape need be [1, channel]
            slope_data = _const_weight_or_none(val_slope)
            slope_data = np.reshape(slope_data, [1] + shape_slope)
            self.weights[val_slope.layer_name] = slope_data

        self.omit_nodes.append(val_slope.layer_name)
        attr = {
            "param_attr": string(val_slope.layer_name),
            'mode': string(mode)
        }
        node.fluid_code.add_layer(
            "prelu", inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def Squeeze(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        axes = node.get_attr('axes')
        attr = {'axes': axes, "name": string(node.layer_name)}
        if len(val_x.out_shapes[0]) == 1:
            node.fluid_code.add_layer(
                "cast",
                inputs=val_x,
                output=node,
                param_attr={'dtype': string(val_x.dtype)})
        else:
            node.fluid_code.add_layer(
                "squeeze", inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def Equal(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_y = self.graph.get_input_node(node, idx=1, copy=True)
        node.fluid_code.add_layer(
            "equal",
            inputs={'x': val_x,
                    'y': val_y},
            output=node,
            param_attr=None)

    @print_mapping_info
    def Greater(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_y = self.graph.get_input_node(node, idx=1, copy=True)
        node.fluid_code.add_layer(
            "greater_than",
            inputs={'x': val_x,
                    'y': val_y},
            output=node,
            param_attr=None)

    @print_mapping_info
    def Where(self, node):
        condition = self.graph.get_input_node(node, idx=0, copy=True)
        val_x = self.graph.get_input_node(node, idx=1, copy=True)
        val_y = self.graph.get_input_node(node, idx=2, copy=True)

        not_condition = condition.layer_name + '_not'
        node.fluid_code.add_layer(
            "logical_not",
            inputs=condition,
            output=not_condition,
            param_attr=None)
        cast_not_condition = not_condition + '_cast'
        node.fluid_code.add_layer(
            "cast",
            inputs=not_condition,
            output=cast_not_condition,
            param_attr={'dtype': string(val_x.dtype)})
        cast_condition = condition.layer_name + '_cast'
        node.fluid_code.add_layer(
            "cast",
            inputs=condition,
            output=cast_condition,
            param_attr={'dtype': string(val_x.dtype)})
        mul_val_x = val_x.layer_name + '_mul'
        node.fluid_code.add_layer(
            "elementwise_mul",
            inputs={'x': val_x,
                    'y': cast_condition},
            output=mul_val_x,
            param_attr=None)
        mul_val_y = val_y.layer_name + '_mul'
        node.fluid_code.add_layer(
            "elementwise_mul",
            inputs={'x': val_y,
                    'y': cast_not_condition},
            output=mul_val_y,
            param_attr=None)

        node.fluid_code.add_layer(
            "elementwise_add",
            inputs={'x': mul_val_x,
                    'y': mul_val_y},
            output=node,
            param_attr=None)

    @print_mapping_info
    def NonZero(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_x_dim = len(val_x.out_shapes[0])
        if val_x_dim == 1:
            node.fluid_code.add_layer("nonzero", inputs=val_x, output=val_x)
            node.fluid_code.add_layer(
                "transpose",
                inputs=val_x,
                output=node,
                param_attr={'perm': [1, 0]})
        if val_x_dim > 1:
            node.fluid_code.add_layer("nonzero", inputs=val_x, output=val_x)
            node.fluid_code.add_layer(
                "split",
                inputs=val_x,
                output=val_x,
                param_attr={'num_or_sections': 1,
                            'dim': val_x_dim})
            node.fluid_code.add_layer("concat", inputs=val_x, output=node)

    @print_mapping_info
    def Identity(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        node.fluid_code.add_layer("assign", inputs=val_x, output=node)

    @print_mapping_info
    def Tile(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_repeats = self.graph.get_input_node(node, idx=1, copy=True)
        repeats = _const_weight_or_none(val_repeats)

        if repeats is None:
            repeats = val_repeats.layer_name
            if val_repeats.dtype != 'int32':
                attr = {"dtype": string("int32")}
                node.fluid_code.add_layer(
                    "cast",
                    inputs=repeats,
                    output="{}.tmp".format(repeats),
                    param_attr=attr)
                repeats = "{}.tmp".format(repeats)

        elif isinstance(repeats, int):
            repeats = [repeats]

        attr = {
            'expand_times': repeats,
            "name": string(node.layer_name),
        }
        node.fluid_code.add_layer(
            "expand", inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def MaxPool(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        auto_pad = node.get_attr('auto_pad', 'NOTSET')
        assert node.get_attr(
            "dilations") is None, 'only dilations = 0 is supported'  # optional

        kernel_shape = node.get_attr("kernel_shape")
        poolnd = len(kernel_shape)
        strides = node.get_attr("strides")
        pad_mode = node.get_attr("pads")
        ceil_mode = bool(node.get_attr('ceil_mode', 0))  # optional
        pads = node.get_attr('pads', [0] * (poolnd * 2))  # optional
        fluid_op = 'pool{}d'.format(poolnd)
        assert 2 <= poolnd <= 3, 'only pool2d and pool3d is supported'

        paddings, val_x = self._pad_if_asymmetric(node, pads, val_x)

        if auto_pad == "SAME_UPPER" or auto_pad == "SAME_LOWER":
            input_shape = val_x.out_shapes[0]
            pad_h = _get_same_padding(input_shape[2], kernel_shape[0],
                                      strides[0])
            pad_w = _get_same_padding(input_shape[3], kernel_shape[1],
                                      strides[1])
            paddings = pad_h + pad_w

        attr = {
            "pool_size": kernel_shape,
            "pool_type": string("max"),
            "pool_stride": strides,
            "pool_padding": paddings,
            "ceil_mode": ceil_mode,
            "name": string(node.layer_name),
            "exclusive": False
        }
        node.fluid_code.add_layer(
            fluid_op, inputs=val_x, output=node, param_attr=attr)

    def _global_pool(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_y = self.graph.get_node(node.layer.output[0], copy=True)
        fluid_op = 'pool2d'
        pool_type = None
        if node.layer.op_type == 'GlobalMaxPool':
            pool_type = 'max'
        elif node.layer.op_type == 'GlobalAveragePool':
            pool_type = 'avg'

        attr = {
            "pool_type": string(pool_type),
            "global_pooling": True,
            "name": string(node.layer_name)
        }
        node.fluid_code.add_layer(
            fluid_op, inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def GlobalMaxPool(self, node):
        self._global_pool(node)

    @print_mapping_info
    def GlobalAveragePool(self, node):
        self._global_pool(node)

    @print_mapping_info
    def Conv(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_w = self.graph.get_input_node(node, idx=1, copy=True)
        val_y = self.graph.get_node(node.layer.output[0], copy=True)

        self.omit_nodes.append(val_w.layer_name)

        has_bias = len(node.layer.input) == 3
        if has_bias:
            val_b = self.graph.get_input_node(node, idx=2, copy=True)
            self.omit_nodes.append(val_b.layer_name)
        auto_pad = node.get_attr('auto_pad', 'NOTSET')

        kernel_shape = node.get_attr('kernel_shape')
        convnd = len(kernel_shape)
        assert 2 <= convnd <= 3, 'only conv2d and conv3d is supported'
        num_out_channels = val_w.out_shapes[0][0]
        fluid_op = 'conv{}d'.format(convnd)

        num_groups = node.get_attr('group', 1)
        strides = node.get_attr('strides', [1] * convnd)
        dilations = node.get_attr('dilations', [1] * convnd)
        pads = node.get_attr('pads', [0] * (convnd * 2))

        input_shape = val_x.out_shapes[0]
        paddings, val_x = self._pad_if_asymmetric(node, pads, val_x)

        if auto_pad == "SAME_UPPER" or auto_pad == "SAME_LOWER":
            pad_h = _get_same_padding(input_shape[2], kernel_shape[0],
                                      strides[0])
            pad_w = _get_same_padding(input_shape[3], kernel_shape[1],
                                      strides[1])
            paddings = pad_h + pad_w

        attr = {
            "num_filters": num_out_channels,
            "filter_size": kernel_shape,
            "stride": strides,
            "padding": paddings,
            "dilation": dilations,
            "groups": num_groups,
            'param_attr': string(val_w.layer_name),
            "name": string(node.layer_name)
        }
        if has_bias:
            attr["bias_attr"] = string(val_b.layer_name)
        else:
            attr["bias_attr"] = False
        node.fluid_code.add_layer(
            fluid_op, inputs=val_x, output=node, param_attr=attr)

    @print_mapping_info
    def ConvTranspose(self, node):
        val_x = self.graph.get_input_node(node, idx=0, copy=True)
        val_w = self.graph.get_input_node(node, idx=1, copy=True)
        val_b = None
        if len(node.layer.input) > 2:
            val_b = self.graph.get_input_node(node, idx=2, copy=True)
            self.omit_nodes.append(val_b.layer_name)
        self.omit_nodes.append(val_w.layer_name)

        val_y = self.graph.get_node(node.layer.output[0], copy=True)

        auto_pad = node.get_attr('auto_pad', 'NOTSET')
        out_padding = node.get_attr('output_padding', [0, 0])
        kernel_shape = node.get_attr('kernel_shape')
        assert kernel_shape, 'kernel_shape not inferred'
        convnd = len(kernel_shape)
        assert 2 <= convnd <= 3, 'only conv2d_transpose and conv3d_transpose supported'
        num_out_channels = val_w.out_shapes[0][1]
        fluid_op = 'conv{}d_transpose'.format(convnd)

        num_groups = node.get_attr('group', 1)
        strides = node.get_attr('strides', [1] * convnd)
        dilations = node.get_attr('dilations', [1] * convnd)
        output_size = node.get_attr('output_shape', [])
        pads = node.get_attr('pads', [0] * (convnd * 2))

        paddings, var_x = self._pad_if_asymmetric(node, pads, val_x)

        output_size = [0, 0]

        output_size[0] = (val_x.out_shapes[0][2] - 1
                          ) * strides[0] - 2 * paddings[0] + dilations[0] * (
                              kernel_shape[0] - 1) + 1 + out_padding[0]
        output_size[1] = (val_x.out_shapes[0][3] - 1
                          ) * strides[1] - 2 * paddings[1] + dilations[1] * (
                              kernel_shape[1] - 1) + 1 + out_padding[1]
        attr = {
            'num_filters': num_out_channels,
            'output_size': output_size or None,
            'filter_size': kernel_shape,
            'padding': paddings,
            'stride': strides,
            'dilation': dilations,
            'groups': num_groups,
            'param_attr': string(val_w.layer_name),
            'bias_attr': None if val_b is None else string(val_b.layer_name),
            'name': string(node.layer_name),
        }
        node.fluid_code.add_layer(
            fluid_op, inputs=val_x, output=node, param_attr=attr)
